# Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any


type JSONValue = (
    dict[str, JSONValue] | list[JSONValue] | str | int | float | bool | None
)


def ensure_dict(value: Any) -> dict[str, Any]:
    if not isinstance(value, dict):
        raise TypeError(f'Expected dict, got {type(value).__name__}')
    return value



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .tool.cli import tatsu_main


if __name__ == '__main__':
    tatsu_main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
col.py ──→ metrics.py
  │           │
  └─── row.py ┘
         │
  multi.py, main.py

Progress bars with terminal animation, multi-bar support, and color styling.
"""

from __future__ import annotations

from .bar import Bar
from .broker import BarBroker
from .col import Col
from .main import main
from .metrics import BarRowProtocol, Metrics
from .multi import Multi
from .row import BarRow, State


__all__ = [
    "BarBroker",
    "BarRow",
    "Bar",
    "BarRowProtocol",
    "Col",
    "main",
    "Metrics",
    "Multi",
    "State",
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
import sysconfig
from collections.abc import Callable, Iterable
from concurrent.futures import as_completed
from itertools import islice
from pickle import PickleError, PicklingError
from typing import Any

from .result import Result
from .task import Event, Func


GIL_DISABLED = sysconfig.get_config_var("Py_GIL_DISABLED")
HAS_MULTITHREADING_SUPPORT = GIL_DISABLED


def active_pmap() -> Callable[
    [Event, Func, Iterable[Any], int | None], Iterable[Result]
]:
    import multiprocessing
    from concurrent.futures import (
        Executor,
        ProcessPoolExecutor,
        ThreadPoolExecutor,
    )

    def executor_pmap(
        executorcls: type[Executor],
        stop: Event,
        process: Func,
        tasks: Iterable[Any],
        max_workers: int | None = None,
    ) -> Iterable[Result]:
        # by Copilot 2026-03-06

        if not tasks:
            return

        with executorcls(max_workers=max_workers) as ex:  # type: ignore
            try:
                n = 1 + (max_workers or 8)
                taskiter = iter(tasks)
                if issubclass(executorcls, ProcessPoolExecutor):
                    futures = {
                        ex.submit(process, task): task for task in islice(taskiter, n)
                    }
                else:
                    futures = {ex.submit(process, task): task for task in taskiter}

                while futures:
                    if stop.is_set():
                        ex.shutdown(wait=False, cancel_futures=True)
                        break

                    for future in as_completed(futures):
                        _task = futures.pop(future)

                        if not stop.is_set():
                            for task in islice(taskiter, 1):
                                new_future = ex.submit(process, task)
                                futures[new_future] = task

                        yield future.result()

                        if stop.is_set():
                            break

            except KeyboardInterrupt:
                stop.set()

                print(file=sys.stderr)
                print(file=sys.stderr)
                print("Wait...", file=sys.stderr)
                sys.stderr.flush()

                ex.shutdown(wait=False, cancel_futures=True)

                print("           ", end="\r", file=sys.stderr)
                sys.stderr.flush()

    def thread_pmap(
        event: Event,
        process: Func,
        tasks: Iterable[Any],
        max_workers: int | None = None,
    ) -> Iterable[Result]:
        yield from executor_pmap(
            ThreadPoolExecutor,
            event,
            process,
            tasks,
            max_workers=max_workers or multiprocessing.cpu_count(),
        )

    def process_pmap(
        event: Event,
        process: Func,
        tasks: Iterable[Any],
        max_workers: int | None = None,
    ) -> Iterable[Result]:
        try:
            yield from executor_pmap(
                ProcessPoolExecutor,
                event,
                process,
                tasks,
                max_workers=max_workers or multiprocessing.cpu_count(),
            )
        except (TypeError, PicklingError, PickleError):
            raise
            # yield from thread_pmap(event, process, tasks, max_workers)
        except KeyboardInterrupt:
            return

    if sys.version_info >= (3, 14):
        from concurrent.futures import (
            InterpreterPoolExecutor,  # pyright: ignore[reportAttributeAccessIssue]
        )

        def interpreter_pmap(
            event: Event,
            process: Func,
            tasks: Iterable[Any],
            max_workers: int | None = None,
        ) -> Iterable[Result]:
            try:
                yield from executor_pmap(
                    InterpreterPoolExecutor,
                    event,
                    process,
                    tasks,
                    max_workers=max_workers or multiprocessing.cpu_count(),
                )
            except (TypeError, PicklingError, PickleError):
                yield from thread_pmap(event, process, tasks, max_workers)

    def imap_pmap(process: Func, tasks: Iterable[Any]) -> Iterable[Result]:
        tasks = list(tasks)
        nworkers = 4 * max(1, multiprocessing.cpu_count())

        count = len(tasks)
        with multiprocessing.Pool(processes=nworkers) as pool:
            yield from pool.imap_unordered(process, tasks)
        if len(tasks) != count:
            raise RuntimeError(
                'number of chunked tasks different %d != %d' % (len(tasks), count),
            )

    if HAS_MULTITHREADING_SUPPORT:
        return thread_pmap
    return process_pmap



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterator
from typing import Any

from ..ctx import Ctx, Func
from .loop import LoopContext


class LoopWithSepContext(LoopContext):
    def __init__(self, ctx: Ctx, plus=True, omitsep: bool = True):
        super().__init__(ctx, plus=plus)
        self.omitsep = omitsep
        self._sep: Func | None = None

    @property
    def sep_func(self) -> Func:
        if self._sep is None:
            raise RuntimeError('.sep not set')
        return self._sep

    def sep(self, func: Func) -> Func:
        if self._sep is not None:
            raise RuntimeError('.sep already set')
        self._sep = func
        return func

    def _inner_iter(self, ctx: Ctx, func: Func) -> Iterator[Any]:
        sep = ctx.isolate(self.sep_func)
        if not self.omitsep:
            yield sep
        yield ctx.isolate(func)

    def parse(self, ctx: Ctx) -> Any:
        if self.plus:
            return ctx.positive_closure(
                self.func, sep=self.sep_func, omitsep=self.omitsep
            )
        else:
            return ctx.closure(self.func, sep=self.sep_func, omitsep=self.omitsep)



from __future__ import annotations

import dataclasses
from collections.abc import Generator, Mapping
from types import SimpleNamespace  # noqa: F401  # pyright: ignore[reportUnusedImport]
from typing import Any, Self

from ..util.abctools import isiter
from ..util.asjson import AsJSONMixin
from ..ztyle import Style


__from_json__class__: dict[str, type] = {}


class Object:
    pass


def dataclass_fields(
    cls: type,
) -> Generator[tuple[str, dataclasses.Field], None, None]:
    if not dataclasses.is_dataclass(cls):
        return
    for parent in cls.mro()[1:]:
        yield from dataclass_fields(parent)
    fields = dataclasses.fields(cls)
    yield from [(f.name, f) for f in fields]


class JSONBase(AsJSONMixin):
    def __init_subclass__(cls: type, **kwargs):
        __from_json__class__[cls.__name__] = cls

    @classmethod
    def __from_json__(cls: type[Self], data: Mapping[str, Any]) -> Self:
        if dataclasses.is_dataclass(cls):
            fieldmap: dict[str, dataclasses.Field] = dict(dataclass_fields(cls))
            initdata = {
                name: value
                for name, value in data.items()
                if (f := fieldmap.get(name)) and f.init
            }
            return cls(**initdata)  # type: ignore

        new = cls.__new__(cls)
        for name, value in data.items():
            if name == "__class__":
                continue
            try:
                setattr(new, name, value)
            except AttributeError:
                if hasattr(cls, name):
                    pass  # a read-only attribute
                else:
                    raise
        return new


def fromjson(obj: Any) -> Any:
    """
    Transform serialized JSON into a Python object.
    """

    def dfs(node: Any) -> Any:  # noqa: PLR0911
        def mapped() -> dict[str, Any]:
            map: dict = node
            return {
                name: dfs(value) for name, value in map.items() if name != "__class__"
            }

        def asobj() -> object:
            return SimpleNamespace(**mapped())
            # NOTE another, more complicated approach:
            # obj = Object()
            # for name, value in mapped().items():
            #     setattr(obj, name, value)
            # return obj

        match node:
            case str():
                if node.startswith(("\\e[", "f{")):
                    return Style.from_raw(node)
                return node
            case int() | float() | bool() | bytes() | bytearray() | complex():
                return node
            case Mapping() as map:
                typename = map.get("__class__", None)
                if not typename:
                    return mapped()
                if (cls := __from_json__class__.get(typename)) is not None:
                    assert issubclass(cls, JSONBase)
                    return cls.__from_json__(mapped())  # NOTE the raw contents
                return asobj()
            case list() | tuple() | set() as seq:
                return [dfs(e) for e in seq]
            case _ if isiter(node):
                return [dfs(e) for e in node]
            case _:
                return node

    return dfs(obj)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from importlib import resources


grammar_path = resources.files().joinpath("_tatsu.ebnf")  # noqa: RUF067
grammar = grammar_path.read_text()  # noqa: RUF067



import argparse


def print_parser_tree(parser, indent=0):
    prefix = "  " * indent
    # Print the name of the parser or command
    print(f"{prefix}Command: {parser.prog}")

    # Print arguments for this level
    for action in parser._actions:
        if not isinstance(action, (argparse._HelpAction, argparse._SubParsersAction)):
            print(f"{prefix}  - {action.option_strings}")

    # Recurse into subparsers
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            for choice, subparser in action.choices.items():
                print(f"{prefix}Subcommand: {choice}")
                print_parser_tree(subparser, indent + indent + 1)


# Usage:
# print_parser_tree(parser)



#!/usr/bin/env python3
# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# noqa # type: ignore # ruff: noqa
from __future__ import annotations

import argparse
import importlib.util
import sys
import time
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path


have_tiexiu: bool = False
try:
    # noinspection PyUnusedImports
    import tiexiu  # type: ignore

    have_tiexiu = True
except ImportError:
    pass

have_ogopego: bool = False
try:
    import ogopego  # type: ignore

    have_ogopego = True
except ImportError:
    pass

from .. import peg
from ..api import compile, to_python_sourcecode
from ..exceptions import FailedParse
from ..parsing import Parser
from ..util.common import try_read, typename
from ..util.strtools import countlines
from ..util.timetools import timer


@dataclass
class BenchmarkResult:
    typename: str
    file_count: int
    lines_parsed: int
    error_count: int
    failed_files: list[str]  # Added this field
    setup_time: float
    total_parsing_time: float
    avg_parsing_time: float
    avg_lines_sec: float


def _setup_mem_parser(grammar_src: str) -> tuple[peg.Grammar, float]:
    with timer() as t:
        model = compile(grammar_src)
        model = model.optimized()
    return model, t.delta


def _setup_gen_parser(
    grammar_src: str,
    grammar_name: str,
) -> tuple[Parser, float, Path]:
    parser_file = f"temp_parser_{int(time.time())}.py"
    parser_path = Path(parser_file).resolve()

    with timer() as tgen:
        python_source = to_python_sourcecode(grammar_src, name=grammar_name)
        parser_path.write_text(python_source, encoding='utf-8')

        spec = importlib.util.spec_from_file_location("temp_gen_parser", parser_path)
        if not (spec and spec.loader):
            raise ImportError("could not create module spec")

        module = importlib.util.module_from_spec(spec)
        sys.modules["temp_gen_parser"] = module
        spec.loader.exec_module(module)

        parser_class = getattr(module, f"{grammar_name}Parser", None)
        if not parser_class:
            for obj in vars(module).values():
                if (
                    isinstance(obj, type)
                    and issubclass(obj, Parser)
                    and obj is not Parser
                ):
                    parser_class = obj
                    break
        if not parser_class:
            raise RuntimeError("could not find a generated parser class in the module.")
        parser = parser_class()

    gen_time = tgen.delta
    return parser, gen_time, parser_path


def _print_run_details(
    title: str,
    result: BenchmarkResult,
    lbl_w: int,
    num_fmt: str,
    verbose: bool,
):
    print(f"\n--- {title} ---")
    if verbose and result.failed_files:
        print(f"{'failed files:\n':<{lbl_w}}")
        print('\n'.join(result.failed_files))
        print()
    print(f"typename: {result.typename}")
    print(f"{'one-time setup:':<{lbl_w}}{num_fmt.format(result.setup_time)} s")
    print(
        f"{f'total parsing time ({result.file_count} files):':<{lbl_w}}"
        f"{num_fmt.format(result.total_parsing_time)} s",
    )
    err_pct = 100 * result.error_count / result.file_count
    print(
        f"{f'errors ({result.error_count}/{result.file_count} files):':<{lbl_w}}"
        f"{num_fmt.format(err_pct)} %",
    )
    print(
        f"{'average parsing time:':<{lbl_w}}{num_fmt.format(result.avg_parsing_time)} s/file",
    )
    print(f"{'average speed:':<{lbl_w}}{num_fmt.format(result.avg_lines_sec)} sloc/sec")


def print_performance_comparison(results: list[tuple[str, BenchmarkResult]]):
    if not results:
        return

    # Map full names to mnemonics and collect speeds
    mnemonic_map = {
        "in-memory": "mem",
        "generated": "gen",
        "tiexiu": "xiu",
        "ogopego": "ogo",
    }

    # Filter out results that don't have a mnemonic mapping
    # and create a list of (mnemonic, speed) tuples
    parsed_results = []
    for name, res in results:
        mnemonic = mnemonic_map.get(name.lower())
        if mnemonic:
            parsed_results.append((mnemonic, res.avg_lines_sec))

    if len(parsed_results) < 2:
        # Not enough results for a comparison table
        return

    # Prepare data for the table
    mnemonics = [r[0] for r in parsed_results]
    speeds = [r[1] for r in parsed_results]

    # Determine column width for numbers and ratios
    # Max length of "mnemonic (N)" for the first column
    max_label_len = max(len(f"{m} ({i + 1})") for i, m in enumerate(mnemonics))

    # Max length for sloc/s values
    max_speed_val_len = max(len(f"{s:.2f}") for s in speeds) if speeds else 0

    # Width for the sloc/s column, including its header
    sloc_s_col_width = max(len("sloc/s"), max_speed_val_len) + 2  # +2 for padding

    # Width for ratio columns (e.g., " 2.00 " or " inf ")
    ratio_col_width = 7  # Minimum width for " X.XX " or "  inf  "

    # The first column will contain "mnemonic (N)"
    # The second column will contain "sloc/s" values
    # Subsequent columns will contain ratios (1...N)

    # Print header row
    # Empty space for the "mnemonic (N)" column
    header = f"{'':<{max_label_len}}"
    # Header for the "sloc/s" column
    header += f"{'sloc/s':>{sloc_s_col_width}}"
    # Headers for the ratio columns (1...N)
    for i in range(len(mnemonics)):
        header += f"{i + 1:>{ratio_col_width}}"
    print(header)

    # Print data rows
    for i, (row_mnemonic, row_speed) in enumerate(parsed_results):
        # First column: mnemonic (N)
        row_label = f"{row_mnemonic} ({i + 1})"
        row_str = f"{row_label:<{max_label_len}}"

        # Second column: sloc/s value
        row_str += f"{row_speed:>{sloc_s_col_width}.2f}"

        # Subsequent columns: ratios
        for j, (_col_mnemonic, col_speed) in enumerate(parsed_results):
            if i == j:
                row_str += f"{'-':>{ratio_col_width}}"
            else:
                # This calculation (row_speed / col_speed) matches the original verbose output's logic
                # and the example output provided by the user.
                if col_speed == 0:
                    ratio_str = "inf"
                else:
                    ratio = row_speed / col_speed
                    ratio_str = f"{ratio:>{ratio_col_width}.2f}"
                row_str += ratio_str
        print(row_str)


def print_summary(
    grammar: str,
    count: int,
    mem_run: BenchmarkResult | None,
    gen_run: BenchmarkResult | None,
    tiexiu_run: BenchmarkResult | None,
    ogo_run: BenchmarkResult | None = None,
    verbose: bool = False,
):
    relgrammar = Path(grammar).resolve().relative_to(Path.cwd())
    print(f"grammar: {relgrammar}")
    print(f"input files: {count}")

    all_runs = [r for r in [mem_run, gen_run, tiexiu_run, ogo_run] if r is not None]
    all_numbers = []
    for r in all_runs:
        all_numbers.extend(
            [
                r.setup_time,
                r.total_parsing_time,
                r.avg_parsing_time,
                r.avg_lines_sec,
            ],
        )

    int_width = max(len(str(int(abs(n)))) for n in all_numbers) if all_numbers else 0
    num_width = int_width + 3
    num_fmt = f"{{:>{num_width}.2f}}"

    count_width = len(str(count))
    longest_label = f"total parsing time ({'9' * count_width} files):"
    labels = [
        "one-time setup:",
        longest_label,
        "average parsing time:",
        "average speed:",
        "in-memory:",
        "generated:",
        "tiexiu:",
        "ogopego:",
        "failed files:",
    ]
    lbl_w = max(len(lbl) for lbl in labels) + 2

    if mem_run:
        _print_run_details("in-memory model", mem_run, lbl_w, num_fmt, verbose)
    if gen_run:
        _print_run_details("generated python parser", gen_run, lbl_w, num_fmt, verbose)
    if tiexiu_run:
        _print_run_details("tiexiu (rust) parser", tiexiu_run, lbl_w, num_fmt, verbose)
    if ogo_run:
        _print_run_details("ogopego (go) parser", ogo_run, lbl_w, num_fmt, verbose)

    comparison_results = []
    if mem_run:
        comparison_results.append(("in-memory", mem_run))
    if gen_run:
        comparison_results.append(("generated", gen_run))
    if tiexiu_run:
        comparison_results.append(("tiexiu", tiexiu_run))
    if ogo_run:
        comparison_results.append(("ogopego", ogo_run))

    if len(comparison_results) > 1:
        print("\n--- comparison (sloc/sec ratios) ---")
        print_performance_comparison(comparison_results)


def benchmark(
    grammar: str | Path,
    filenames: Iterable[str | Path],
    mode: set[str],
) -> tuple[
    BenchmarkResult | None,
    BenchmarkResult | None,
    BenchmarkResult | None,
    BenchmarkResult | None,
]:
    oldlimit = sys.getrecursionlimit()
    sys.setrecursionlimit(2**16)
    try:
        grampath = Path(grammar)
        gramsrc = grampath.read_text(encoding="utf-8")

        model, memsetup = _setup_mem_parser(gramsrc)
        gramname = model.name or "Benchmark"

        currentpath = Path().absolute()
        filepaths = [Path(f).relative_to(currentpath, walk_up=True) for f in filenames]
        texts = [try_read(p) for p in filepaths]
        nfiles = len(texts)

        memrun = None
        if 'mem' in mode:
            memtime = 0.0
            memerrs = 0
            lines_parsed = 0
            mem_failed: list[str] = []
            for i, text in enumerate(texts):
                pct = int((i + 1) / nfiles * 100)
                print(f"[Mem {pct:3d}%] Benchmarking memory parser...", end="\r")
                with timer() as t:
                    try:
                        model.parse(text, asmodel=True)
                        lines_parsed += countlines(text).code
                    except FailedParse:
                        memerrs += 1
                        mem_failed.append(str(filepaths[i]))
                    memtime += t.delta
            memrun = BenchmarkResult(
                typename(model),
                file_count=nfiles,
                error_count=memerrs,
                lines_parsed=lines_parsed,
                failed_files=mem_failed,
                setup_time=memsetup,
                total_parsing_time=memtime,
                avg_parsing_time=memtime / nfiles if nfiles else 0,
                avg_lines_sec=lines_parsed / memtime if memtime else 0,
            )

        # --- Loop 2: Generated Parser ---
        genrun = None
        if 'gen' in mode:
            parser, gensetup, parserpath = _setup_gen_parser(gramsrc, gramname)
            gensetup += memsetup  # Account for initial grammar compilation
            gentime = 0.0
            generrs = 0
            gen_failed: list[str] = []
            lines_parsed = 0
            for i, text in enumerate(texts):
                pct = int((i + 1) / nfiles * 100)
                print(f"[Gen {pct:3d}%] Benchmarking generated parser...", end="\r")
                with timer() as t:
                    try:
                        parser.parse(text, asmodel=True)
                        lines_parsed += countlines(text).code
                    except FailedParse:
                        generrs += 1
                        gen_failed.append(str(filepaths[i]))
                    gentime += t.delta
            genrun = BenchmarkResult(
                typename(parser),
                file_count=nfiles,
                error_count=generrs,
                lines_parsed=lines_parsed,
                failed_files=gen_failed,
                setup_time=gensetup,
                total_parsing_time=gentime,
                avg_parsing_time=gentime / nfiles if nfiles else 0,
                avg_lines_sec=lines_parsed / gentime if gentime else 0,
            )
            parserpath.unlink()

        tiexiu_run = None
        if 'tiexiu' in mode and have_tiexiu:
            tiexiu_time = 0.0
            tiexiu_errs = 0
            lines_parsed = 0
            tiexiu_failed: list[str] = []
            # Tiexiu setup is basically zero (it parses the grammar on every call currently)
            # or it might have a one-time overhead for the first call
            peg = tiexiu.pegapi()  # type: ignore
            with timer() as t:
                peg.compile(gramsrc)
                tiesetup = t.delta
            for i, text in enumerate(texts):
                pct = int((i + 1) / nfiles * 100)
                print(f"[Xiu {pct:3d}%] Benchmarking tiexiu parser...", end="\r")
                with timer() as t:
                    try:
                        peg.parse_to_json_string(
                            gramsrc, text, source=str(filepaths[i])
                        )
                        lines_parsed += countlines(text).code
                    except ValueError as _e:
                        tiexiu_errs += 1
                        # tiexiu_failed.append(f"{filepaths[i]} (Error: {e})")
                        tiexiu_failed.append(str(filepaths[i]))
                    except Exception as _e:  # Catching other unexpected errors
                        tiexiu_errs += 1
                        # tiexiu_failed.append(
                        #     f"{filepaths[i]} (Unexpected: {type(e).__name__}: {e})",
                        # )
                        tiexiu_failed.append(str(filepaths[i]))
                    tiexiu_time += t.delta
            tiexiu_run = BenchmarkResult(
                "tiexiu.parse",
                file_count=nfiles,
                error_count=tiexiu_errs,
                lines_parsed=lines_parsed,
                failed_files=tiexiu_failed,
                setup_time=tiesetup,
                total_parsing_time=tiexiu_time,
                avg_parsing_time=tiexiu_time / nfiles if nfiles else 0,
                avg_lines_sec=lines_parsed / tiexiu_time if tiexiu_time else 0,
            )

        ogo_run = None
        if 'ogo' in mode and have_ogopego:
            ogotime = 0.0
            ogoerrs = 0
            lines_parsed = 0
            ogo_failed: list[str] = []
            with timer() as t:
                ogopego.compile(gramsrc)
                ogosetup = t.delta
            for i, text in enumerate(texts):
                pct = int((i + 1) / nfiles * 100)
                print(f"[Ogo {pct:3d}%] Benchmarking ogopego parser...", end="\r")
                with timer() as t:
                    try:
                        ogopego.parse(gramsrc, text)
                        lines_parsed += countlines(text).code
                    except ogopego.OgoError:
                        ogoerrs += 1
                        ogo_failed.append(str(filepaths[i]))
                    except Exception as _e:
                        ogoerrs += 1
                        ogo_failed.append(str(filepaths[i]))
                    ogotime += t.delta
            ogo_run = BenchmarkResult(
                "ogopego.parse",
                file_count=nfiles,
                error_count=ogoerrs,
                lines_parsed=lines_parsed,
                failed_files=ogo_failed,
                setup_time=ogosetup,
                total_parsing_time=ogotime,
                avg_parsing_time=ogotime / nfiles if nfiles else 0,
                avg_lines_sec=lines_parsed / ogotime if ogotime else 0,
            )

        print(" " * 75)  # Clear the status line

        return memrun, genrun, tiexiu_run, ogo_run
    finally:
        sys.setrecursionlimit(oldlimit)


def add_argparse_options(parser: argparse.ArgumentParser) -> None:
    mode_group = parser
    mode_group.add_argument(
        '--all',
        help='benchmark all types of parser (default)',
        action='store_true',
    )
    mode_group.add_argument(
        '--both',
        help='benchmark mem and gen parsers only (default: all)',
        action='store_true',
    )
    mode_group.add_argument(
        '--mem',
        help='benchmark in-memory parser',
        action='store_true',
    )
    mode_group.add_argument(
        '--gen',
        help='benchmark generated parser',
        action='store_true',
    )
    mode_group.add_argument(
        '--tiexiu',
        help='benchmark tiexiu parser',
        action='store_true',
    )
    mode_group.add_argument(
        '--ogo',
        help='benchmark ogopego parser',
        action='store_true',
    )
    if '--verbose' not in parser._option_string_actions:
        parser.add_argument(
            '--verbose',
            help='show error output from failed parses',
            action='store_true',
        )
    parser.add_argument("grammar", type=Path, help="path to the grammar file")
    parser.add_argument(
        "inputs",
        type=Path,
        nargs='+',
        help="path to one or more input text files",
    )


def add_bench_cmd(sub) -> argparse.ArgumentParser:
    parser = sub.add_parser(
        "bench",
        help="Run benchmarks using known parser implementations",
    )
    add_argparse_options(parser)
    return parser


def bench_cmd(parser: argparse.ArgumentParser) -> int:
    add_argparse_options(parser)
    # mode = parser.add_mutually_exclusive_group()
    args = parser.parse_args()

    for p in [args.grammar, *args.inputs]:
        if not p.exists():
            print(f"error: file '{p}' not found.")
            return 1
    mode = set()
    all = True
    if args.mem or args.both or args.all:
        mode |= {'mem'}
        all = False
    if args.gen or args.both or args.all:
        mode |= {'gen'}
        all = False
    if args.tiexiu or args.all:
        mode |= {'tiexiu'}
        all = False
    if args.ogo or args.all:
        mode |= {'ogo'}
        all = False
    if all:
        mode = {'mem', 'gen', 'tiexiu', 'ogo'}

    if {'tiexiu'} in mode and not have_tiexiu:
        print("warning: tiexiu not found, skipping tiexiu benchmark.")
        if mode == {'tiexiu'}:
            return 1

    if 'ogo' in mode and not have_ogopego:
        print("warning: ogopego not found, skipping ogopego benchmark.")
        if mode == {'ogo'}:
            return 1

    try:
        grammar_path = args.grammar.resolve()
        input_paths = [p.resolve() for p in args.inputs]
        mem_run, gen_run, tiexiu_run, ogo_run = benchmark(
            grammar_path,
            input_paths,
            mode=mode,
        )
        print_summary(
            str(args.grammar),
            len(args.inputs),
            mem_run,
            gen_run,
            tiexiu_run,
            ogo_run=ogo_run,
            verbose=args.verbose,
        )
    except Exception as e:
        print(f"\nan error occurred: {e}", file=sys.stderr)
        raise

    return 0


def main() -> int:
    parser = argparse.ArgumentParser(description="benchmark tatsu parsing methods")
    return bench_cmd(parser)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""Multi-bar display manager with a background rendering thread."""

from __future__ import annotations

import copy
import itertools
import multiprocessing
import shutil
import sys
import threading
import time
from typing import Any, TextIO

from ..util.debugging import prints
from ..util.primality import primes_upto
from ..util.tty import (
    blankpad,
    hide_cursor,
    pushup,
    shoot_lines,
    show_cursor,
)
from ..ztyle import visual_len as vlen
from .row import BarRow


MAXW = shutil.get_terminal_size().columns - 1
MAXL = shutil.get_terminal_size().lines - 1


_screen_lock = multiprocessing.Lock()


class MessageRow(BarRow):
    def is_active(self) -> bool:
        # NOTE Lie to always get rendered
        return True

    def has_started(self) -> bool:
        return True

    def is_stopping(self) -> bool:
        return False

    def is_stopped(self) -> bool:
        return False


class Multi:
    def __init__(
        self,
        rows: list[BarRow],
        /,
        fps: int = 12,  # noqa: B006
        out: TextIO = sys.stderr,
    ):
        self.rows = rows[:]
        self.fps = fps
        self.out = out

        self.alive = False
        self.lock = threading.RLock()
        self.worker = None

        self.height: int = -1
        self.msg_count: int = 0
        self._frame: list[str] = []

    def find_row(self, id: str) -> BarRow | None:
        for r in self.rows:
            if r.id == id:
                return r
        return None

    def remove_row(self, id: str) -> None:
        with self.lock:
            self.rows = [row for row in self.rows if row.id != id]

    def add_row(self, row: BarRow) -> None:
        """Stores a row internally."""
        if existing := self.find_row(row.id):
            existing.update(**row.snap())
            return
        with self.lock:
            self.rows = [*self.rows, row]

    def insert_row(self, index: int, row: BarRow) -> None:
        """Inserts a row at the given index."""
        if existing := self.find_row(row.id):
            existing.update(**row.snap())
            return
        with self.lock:
            self.rows = [*self.rows[:index], row, *self.rows[index:]]

    def insert_message(self, msg: str) -> None:
        """Inserts a message row at the bottom."""
        row = MessageRow(cols=[msg])
        with self.lock:
            self.insert_row(self.msg_count, row)
            self.msg_count += 1

    def update_row(self, snap: dict[str, Any]) -> None:
        """Inserts or updates a row internally."""
        if (id := snap.get("id")) is None:
            return

        with self.lock:
            row = self.find_row(id)
            if row is None:
                return
            row.update(**snap)

    def print(self, *args, **kwargs) -> None:
        """Prints to the output stream."""
        kwargs.pop("file", None)
        kwargs.pop("highlight", None)
        kwargs.pop("markup", None)
        s = prints(*args, end="", **kwargs)
        if not s:
            self.insert_message("")
        for msg in s.splitlines():
            self.insert_message(msg)

    def start(self):
        """Starts the completely isolated background rendering thread."""
        self.alive = True

        self.worker = threading.Thread(target=self._render_loop, daemon=True)
        self.worker.start()

    def stop(self):
        """Gracefully shuts down the rendering thread."""
        with self.lock:
            for row in self.rows:
                row.stop()
                row.stop()

        self.alive = False
        if self.worker:
            self.worker.join()

    def _render_loop(self):
        """The isolated thread handler. Read-only side."""
        self.out.write(hide_cursor())  # Hide cursor
        self.out.write("\n")  # Hide cursor
        self.out.flush()
        fpscycle = itertools.cycle(primes_upto(self.fps, 11))
        try:
            while self.alive:
                self.paint_frame()
                time.sleep(1 / next(fpscycle))
        finally:
            self.out.write(show_cursor())  # Show cursor
            self.out.flush()
        self.paint_frame(final=True)

    def _take_snapshot(self) -> list[BarRow]:
        with self.lock:
            snapshot = [copy.copy(r) for r in self.rows if r.is_active()][-MAXL:]
            # self.rows = [r for r in self.rows if not r.is_stopped()][-MAXL:]
            # self.msg_count = sum(
            #     1
            #     for _ in itertools.takewhile(
            #         lambda r: isinstance(r, MessageRow), self.rows
            #     )
            # )
            return snapshot

    def render_frame(self, *, final: bool = False) -> list[str]:
        snapshot: list[BarRow] = self._take_snapshot()
        rendered: list[list[Any]] = [b._call_render() for b in snapshot]

        colwidth = [self.line_col_widths(line, MAXW) for line in rendered]
        assert len(colwidth) == len(rendered)

        expanded: list[str] = [
            self.expand_row(line, cw) for line, cw in zip(rendered, colwidth)
        ]

        shot_lines = shoot_lines(expanded)

        h = len(expanded)
        c = max(0, self.height - h) if not final else 0
        blank_lines = blankpad(c)

        frame = [*shot_lines, *blank_lines][-MAXL:]

        return frame

    def paint_frame(self, *, final: bool = False) -> None:
        frame = self.render_frame()
        screenshot: str = "".join(frame)

        h = max(len(frame), self.height)
        with _screen_lock:
            self.out.write(screenshot)
            if not final and self.height:
                self.out.write(pushup(len(frame)))
            self.out.flush()
        self.height = h

    def expand_row(self, line: list[Any], cw: list[int]) -> str:
        return ''.join(f"{col:{w}}" for col, w in zip(line, cw))

    def line_col_widths(self, line: list[Any], maxw: int) -> list[int]:
        def colwidth(col: Any) -> int:
            match col:
                case str() as s:
                    return vlen(s)
                case _:
                    return len(col)

        cw = [0] * len(line)
        fill: set[int] = set()
        for j, col in enumerate(line):
            w = colwidth(col)
            if w <= 0:
                fill.add(j)
            else:
                cw[j] = w

        budget = max(0, maxw - sum(cw))
        while fill and budget > 0:
            count = len(fill)
            w = round(((1 / count) + budget) / count)
            budget -= w
            j = fill.pop()
            cw[j] = w

        return cw



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterable
from functools import cache
from typing import Any, override

from ..util import make_hashable, typename
from ..util.asjson import asjson
from .cst import cstadd, cstaddlist
from .infos import ParseInfo


class AST(dict[str, Any]):
    """
    A dictionary that allows attribute-style access to its keys.
    # by Gemini (2026-01-26)
    # by [apalala@gmail.com](https://github.com/apalala)
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__()
        self.__dict__['__frozen__'] = False
        self.update(*args, **kwargs)
        self.__dict__['__frozen__'] = True

    @property
    def parseinfo(self) -> ParseInfo | None:
        return self.get('__parseinfo__')

    # NOTE: required to bypass '__frozen__'
    def set_parseinfo(self, value: ParseInfo | None) -> None:
        super().__setitem__('parseinfo', value)
        super().__setitem__('__parseinfo__', value)

    def asjson(self) -> Any:
        return asjson(self)

    def _set(self, key: str, node: Any) -> None:
        key = self._safekey(key)
        cst = self.get(key)
        super().__setitem__(key, cstadd(cst, node))

    def _setlist(self, key: str, node: Any) -> None:
        key = self._safekey(key)
        cst = self.get(key)
        super().__setitem__(key, cstaddlist(cst, node))

    @staticmethod
    @cache
    def _unsafe() -> frozenset[str]:
        return frozenset(vars(dict).keys())

    def _safekey(self, key: str) -> str:
        while key in self._unsafe():
            key += '_'
        return key

    def _define(self, keys: Iterable[str], list_keys: Iterable[str] | None = None):
        for key in (self._safekey(k) for k in list_keys or []):
            if key not in self:
                super().__setitem__(key, [])
        for key in (self._safekey(k) for k in keys):
            if key not in self:
                super().__setitem__(key, None)

    def __getitem__(self, key: str) -> Any:
        if key in self:
            return super().__getitem__(key)
        return super().get(self._safekey(key))

    def __setitem__(self, key: Any, value: Any) -> None:
        self._set(key, value)

    def __delitem__(self, key: Any) -> None:
        super().__delitem__(self._safekey(key))

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith('_'):
            self.__dict__[name] = value
            return
        # HACK AHEAD'!
        #   allow some attributes for backwards compatibility with legacy
        if name != 'linecount' and self.__dict__.get('__frozen__'):
            raise AttributeError(
                f'{typename(self)} attributes are frozen. Cannot set "{name}".'
            )
        self[name] = value

    def __getattr__(self, name: str) -> Any:
        return self[name]

    def __reduce__(self) -> tuple[Any, Any]:
        return AST, (tuple(self.items()),)

    def __str__(self) -> str:
        return str(self.asjson())

    @override
    def __hash__(self) -> int:  # type: ignore
        return hash(make_hashable(self))



import copy
import sys
import threading
import time


class Bar:
    """A rich, lightweight, fully picklable data object given to the user."""

    def __init__(self, label: str, total: int = 100):
        self.label = label
        self.total = total
        self.current = 0

    def advance(self, delta: int = 1):
        """Write-only operation from the user's side."""
        self.current = min(self.current + delta, self.total)

    def update(self, value: int):
        """Write-only operation from the user's side."""
        self.current = min(value, self.total)


class MultiProgress:
    def __init__(self):
        self._bars = []
        self._lock = threading.Lock()
        self._running = False
        self._thread = None

    def add_bar(self, label: str, total: int = 100) -> Bar:
        """Creates a bar, stores it internally, and returns it to the user."""
        bar = Bar(label, total)
        with self._lock:
            self._bars.append(bar)
        return bar

    def start(self):
        """Starts the completely isolated background rendering thread."""
        self._running = True
        sys.stdout.write("\033[?25l")  # Hide cursor
        sys.stdout.flush()

        self._thread = threading.Thread(target=self._render_loop, daemon=True)
        self._thread.start()

    def stop(self):
        """Gracefully shuts down the rendering thread."""
        self._running = False
        if self._thread:
            self._thread.join()
        # Clean up terminal lines
        with self._lock:
            num_bars = len(self._bars)
        sys.stdout.write(f"\033[{num_bars}B\n\033[?25h")
        sys.stdout.flush()

    def _render_loop(self):
        """The isolated thread handler. Read-only side."""
        while self._running:
            # 1. LOCK INSTANTLY: Take a snapshot copy of the data objects
            with self._lock:
                snapshot = copy.deepcopy(self._bars)

            # 2. RELEASED LOCK: Do the slow I/O printing out-of-band
            if not snapshot:
                time.sleep(0.05)
                continue

            lines = []
            for b in snapshot:
                pct = int((b.current / b.total) * 100) if b.total > 0 else 0
                fill = int(pct / 5)
                lines.append(f"{b.label}: [{'█' * fill}{'-' * (20 - fill)}] {pct}%")

            sys.stdout.write("\n".join(lines))
            sys.stdout.flush()

            time.sleep(0.05)  # 20 FPS
            sys.stdout.write(f"\033[{len(snapshot)}A")  # Reset cursor up



import unittest

from tatsu import g2e, peg as g


class G2ETests(unittest.TestCase):
    def test_simple_parser_rule(self):
        model = g2e.translate(text="grammar T; start: 'hello';", name='T')
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_token_ref_after_definition(self):
        model = g2e.translate(text="grammar T; INT: [0-9]+; start: INT;", name='T')
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('INT', model.rulemap)
        self.assertIn('start', model.rulemap)

    def test_token_ref_before_definition(self):
        model = g2e.translate(text="grammar T; start: INT; INT: [0-9]+;", name='T')
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('INT', model.rulemap)
        self.assertIn('start', model.rulemap)

    def test_undefined_token_becomes_fail(self):
        model = g2e.translate(text="grammar T; start: INT;", name='T')
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('INT', model.rulemap)
        self.assertIsInstance(model.rulemap['INT'].exp, g.Fail)

    def test_camel_case_conversion(self):
        model = g2e.translate(
            text="grammar T; parseExpression: 'expr'; parseStatement: 'stmt';",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('parse_expression', model.rulemap)
        self.assertIn('parse_statement', model.rulemap)

    def test_tokens_block_unreferenced(self):
        model = g2e.translate(
            text="grammar T; tokens { DUMMY } start: 'x';",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_tokens_block_with_reference(self):
        model = g2e.translate(
            text="grammar T; tokens { INT } start: INT;",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('INT', model.rulemap)
        # `tokens { INT }` creates a Void() rule, taking precedence over Fail
        self.assertIsInstance(model.rulemap['INT'].exp, g.Void)

    def test_multiple_alternatives(self):
        model = g2e.translate(
            text="grammar T; start: 'a' | 'b' | 'c';",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_sequence_and_choice(self):
        model = g2e.translate(
            text="grammar T; start: 'a' 'b' ('c' | 'd');",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_optional_and_closure(self):
        model = g2e.translate(
            text="grammar T; start: 'a' ('b')* 'c'?;",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_pretty_output_is_valid(self):
        model = g2e.translate(
            text="grammar T; INT: [0-9]+; start: INT;",
            name='T',
        )
        pretty = model.pretty()
        self.assertIn('INT', pretty)
        self.assertIn('start', pretty)

    def test_grammar_name(self):
        model = g2e.translate(text="grammar Test; start: 'x';", name='Test')
        self.assertEqual(model.name, 'Test')

    def test_named_elements(self):
        model = g2e.translate(
            text="grammar T; start: name='x';",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_negative_lookahead(self):
        model = g2e.translate(
            text="grammar T; start: ~'x' 'y';",
            name='T',
        )
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_empty_grammar(self):
        model = g2e.translate(text="grammar T; start: ;", name='T')
        self.assertIsInstance(model, g.Grammar)
        self.assertIn('start', model.rulemap)

    def test_token_rule_becomes_fail(self):
        model = g2e.translate(
            text="grammar T; INT: [0-9]+;",
            name='T',
        )
        self.assertIsInstance(model.rulemap['INT'].exp, g.Fail)

    def test_fragment_rule_becomes_fail(self):
        model = g2e.translate(
            text="grammar T; fragment DIGIT: [0-9]; fragment LETTER: [a-z]; ID: (LETTER|DIGIT)+;",
            name='T',
        )
        self.assertIsInstance(model.rulemap['DIGIT'].exp, g.Fail)
        self.assertIsInstance(model.rulemap['LETTER'].exp, g.Fail)
        self.assertIsInstance(model.rulemap['ID'].exp, g.Fail)

    def test_token_rule_with_alternatives_becomes_fail(self):
        model = g2e.translate(
            text="grammar T; LETTER: [a-z] | [A-Z] | '_';",
            name='T',
        )
        self.assertIsInstance(model.rulemap['LETTER'].exp, g.Fail)

    def test_forward_ref_token_rule_becomes_fail(self):
        model = g2e.translate(
            text="grammar T; ID: STD_CONTINUE* STD_START; fragment STD_START: [a-z] | '_'; fragment STD_CONTINUE: [a-z] | [0-9];",
            name='T',
        )
        self.assertIsInstance(model.rulemap['ID'].exp, g.Fail)
        self.assertIsInstance(model.rulemap['STD_START'].exp, g.Fail)
        self.assertIsInstance(model.rulemap['STD_CONTINUE'].exp, g.Fail)

    def test_token_rule_unicode_escapes(self):
        model = g2e.translate(
            text="grammar T; UNI: '\\u00AA' | '\\u00B5';",
            name='T',
        )
        self.assertIsInstance(model.rulemap['UNI'].exp, g.Fail)

    def test_token_rule_positive_closure(self):
        model = g2e.translate(
            text="grammar T; fragment DIGIT: [0-9]; FLOAT: DIGIT+ '.' DIGIT+;",
            name='T',
        )
        self.assertIsInstance(model.rulemap['DIGIT'].exp, g.Fail)
        self.assertIsInstance(model.rulemap['FLOAT'].exp, g.Fail)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest

import tatsu
from tatsu.api import to_python_sourcecode
from tatsu.boot import TatSuParserGenerator


CALC_GRAMMAR = r"""
    @@grammar::CALC

    start: expression $

    expression:
        | expression '+' term
        | expression '-' term
        | term

    term:
        | term '*' factor
        | term '/' factor
        | factor

    factor:
        | '(' expression ')'
        | number

    number: /\d+/
"""


@pytest.mark.benchmark
def test_bench_compile_calc_grammar():
    """Benchmark compiling a simple calculator grammar."""
    tatsu.compile(CALC_GRAMMAR)


@pytest.mark.benchmark
def test_bench_parse_arithmetic_expression():
    """Benchmark parsing an arithmetic expression with a compiled grammar."""
    parser = tatsu.compile(CALC_GRAMMAR)
    parser.parse("3 + 5 * ( 10 - 20 )")


@pytest.mark.benchmark
def test_bench_compile_tatsu_grammar():
    """Benchmark compiling the full TatSu grammar (self-hosting)."""
    text = tatsu.grammar
    g = TatSuParserGenerator("TatSuBootstrap")
    g.parse(text)


@pytest.mark.benchmark
def test_bench_codegen_calc_grammar():
    """Benchmark generating Python source code from a grammar."""
    to_python_sourcecode(CALC_GRAMMAR)


@pytest.mark.benchmark
def test_bench_parse_complex_expression():
    """Benchmark parsing a more complex nested expression."""
    parser = tatsu.compile(CALC_GRAMMAR)
    parser.parse("((1 + 2) * (3 + 4)) + ((5 - 6) * (7 + 8)) - 9 * (10 + 11)")



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from os import linesep


LS_LEN = len(linesep)


def take_non_newline_whitespace_len(text: str, start: int = 0) -> int:
    if start >= len(text):
        return 0

    i = start
    while i < len(text):
        c = text[i]
        # beware of multi-character os.linesep
        if not c.isspace() or text.startswith(linesep, i):
            break
        i += 1

    return i - start


def take_linebreak_len(text: str, start: int = 0) -> int | None:
    """Matches a whitespace-only line or EOT without slicing."""
    n = len(text)
    if start >= n:
        return 0

    newline_pos = text.find(linesep, start)
    # The segment to check is from start to newline_pos (exclusive) or to end of text
    end_of_line = newline_pos if newline_pos != -1 else n

    # Efficiency: use a loop to avoid slicing/copying
    for i in range(start, end_of_line):
        if not text[i].isspace():
            return None

    if newline_pos != -1:
        return newline_pos - start + LS_LEN
    return n - start


def take_blankline_len(text: str, start: int = 0) -> int | None:
    """Matches two consecutive whitespace-only lines."""
    off1 = take_linebreak_len(text, start)
    if off1 is None:
        return None

    off2 = take_linebreak_len(text, start + off1)
    if off2 is None:
        return None

    return off1 + off2


def indent_len(text: str, start: int = 0) -> int | None:
    """Returns the length of the leading whitespace of the line at `start`.
    Returns None if the line is empty or contains only whitespace (including the newline).
    """
    n = len(text)
    if start >= n:
        return None

    newline_pos = text.find(linesep, start)
    # Original behavior included the newline in the lstrip() check, so we search up to it
    search_limit = newline_pos + LS_LEN if newline_pos != -1 else n

    for i in range(start, search_limit):
        if not text[i].isspace():
            return i - start

    return None


def take_indent_len(text: str, start: int = 0) -> int | None:
    """Matches a line break followed by an indented line."""
    offset = take_linebreak_len(text, start)
    if offset is None:
        return None

    indent = indent_len(text, start + offset)
    if indent is None:
        return None

    return offset + indent


def take_dedent_len(text: str, pos: int = 0) -> int | None:
    """Matches a line break that returns to the zero-margin."""
    offset = take_linebreak_len(text, pos)
    if offset is None:
        return None

    # Matches if the next line has content starting at the zero margin
    if indent_len(text, pos + offset) == 0:
        return offset
    return None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import tatsu
from tatsu import peg as g


def test_parse_ebnf():
    grammar = r"""
        /*
            Example of a grammar that mixes TatSu and EBNF
        */
        @@grammar :: TatSu  // this is TatSu wiht an EBNF comment

        start := expression $

        expression := expression '+' term | expression '-' term | term

        term := term '*' factor | term '/' factor | factor

        factor := '(' expression ')' | number

        number := /\d+/
    """

    model = tatsu.compile_to_parser(grammar)
    assert isinstance(model, g.Grammar)


def test_optional():
    grammar = r"""
        start:  '[' /abc/

        other := 'xyz'?
    """

    model = tatsu.compile_to_parser(grammar)
    exp = model.rulemap['start'].exp
    assert isinstance(exp, g.Sequence)
    assert repr(exp.sequence) == "[Token('['), Pattern('abc')]"

    exp = model.rulemap['other'].exp
    assert repr(exp) == "Optional(Token('xyz'))"


def test_one_line_grammar():
    grammar = r"""
            start: lisp

            lisp: sexp | list | symbol

            sexp[SExp]: '(' cons=lisp '.' ~ cdr=lisp ')'

            list[List]: '(' ={lisp} ')'

            symbol[Symbol]: /[^\s().]+/

        """

    one_line_grammar = ' ; '.join(s for s in grammar.splitlines() if s.strip())
    parser = tatsu.compile(one_line_grammar)
    model = parser.parse("( abc (x . y))", asmodel=True)
    # FIXME: the order of arguments should be stable in NodeBase.__repr__
    assert repr(model) in {
        "List(Symbol('abc'), SExp(cons=Symbol('x'), cdr=Symbol('y')))",
        "List(Symbol('abc'), SExp(cdr=Symbol('y'), cons=Symbol('x')))",
    }



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Callable
from typing import cast

from ..infos import RuleLike


def leftrec(impl: Callable) -> Callable:
    over: RuleLike = cast(RuleLike, impl)
    over.is_lrec = True
    over.is_memo = False
    return impl


def nomemo(impl: Callable) -> Callable:
    over: RuleLike = cast(RuleLike, impl)
    over.no_memo = True
    return impl


def nostak(impl: Callable) -> Callable:
    over: RuleLike = cast(RuleLike, impl)
    over.no_stak = True
    return impl


def isname(impl: Callable) -> Callable:
    over: RuleLike = cast(RuleLike, impl)
    over.is_name = True
    return impl


def name(impl: Callable) -> Callable:
    over: RuleLike = cast(RuleLike, impl)
    over.is_name = True
    return impl


def token(impl: Callable) -> Callable:
    over: RuleLike = cast(RuleLike, impl)
    over.is_tokn = True
    return impl



# Copyright © 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause

from __future__ import annotations

import importlib.metadata

from .util import Version


__toolname__ = 'TatSu'
__version__ = 'v5.24.1-b.1'
try:
    __toolname__ = importlib.metadata.metadata("TatSu")["name"]
    __version__ = importlib.metadata.version("TatSu")
except importlib.metadata.PackageNotFoundError:
    pass

version = __version__
version_info = Version.parse(version).astuple()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import contextlib
import inspect
import os
import re
from collections import namedtuple
from collections.abc import Callable
from copy import copy
from typing import Any, Self

from ..util.tty import ANSI_RE, SGR_RE, tty_escape, tty_unescape, visual_len


class RGB(namedtuple('RGB', ['r', 'g', 'b'])):
    """A 24-bit RGB color tuple with 0-255 byte clamping.

    Usage:
        RGB(255, 0, 0)  # pure red
        rgb(128, 128, 128)  # via the rgb() helper
    """

    __slots__ = ()

    def __new__(cls, r: int, g: int, b: int) -> Self:
        def to_byte(v: int) -> int:
            return max(0, min(v, 255))

        return super().__new__(cls, to_byte(r), to_byte(g), to_byte(b))


def rgb(r: int, g: int, b: int) -> RGB:
    """Create an RGB tuple from 0-255 byte values."""
    return RGB(r, g, b)


class Color:
    """Policy object controlling whether ANSI color codes are emitted.

    Color evaluates three sources in priority order:
        1. An explicit enable override (set via the constructor or ``.enable()``).
        2. The ``NO_COLOR`` environment variable (disables color when set).
        3. The ``FORCE_COLOR`` environment variable (enables color when set).
        4. Whether ``sys.stdout`` is a TTY (the default heuristic).

    Factory methods:
        ``Color.tty()`` — deference to the TTY check (the default).
        ``Color.always()`` — always emit color.
        ``Color.never()`` — never emit color.
        ``Color.default()`` — same as ``Color()``; self-documenting alias.

    Use ``Color.style(...)`` to create a ``Style`` bound to this policy,
    or pass as ``color=`` to the ``Style`` constructor.
    """

    def __init__(self, enable: bool | None = None):
        self._force_enable: bool | None = enable
        self._check_stderr: bool = False

    def enable(self, value: bool) -> None:
        """Override the enable policy with an explicit boolean.

        After calling ``.enable(False)``, all ``Style`` instances using this
        ``Color`` will return plain text.  Useful to turn color off globally
        based on a CLI flag:

            Color.default().enable(args.no_color)
        """
        self._force_enable = value

    def markup(self, text: str):
        from .markup import markup

        return markup(text, color=self)

    def style(
        self,
        value: str = "",
        *,
        fg: int | RGB | None = -1,
        bg: int | RGB | None = -1,
        bold: bool = False,
        dim: bool = False,
        italic: bool = False,
        underline: bool = False,
        blink: bool = False,
        inverse: bool = False,
        hidden: bool = False,
        strikethrough: bool = False,
    ) -> Style:
        """Create a ``Style`` bound to this color policy.

        All keyword arguments are forwarded to ``Style.__init__``.
        """
        return Style(
            value,
            fg=fg,
            bg=bg,
            bold=bold,
            dim=dim,
            italic=italic,
            underline=underline,
            blink=blink,
            inverse=inverse,
            hidden=hidden,
            strikethrough=strikethrough,
            color=self,
        )

    @classmethod
    def tty(cls) -> Color:
        """Create a Color that defers to TTY detection (default)."""
        return cls()

    @classmethod
    def always(cls) -> Color:
        """Create a Color that always emits ANSI codes."""
        return cls(enable=True)

    @classmethod
    def never(cls) -> Color:
        """Create a Color that never emits ANSI codes."""
        return cls(enable=False)

    @classmethod
    def default(cls) -> Color:
        """Return the system default Color (TTY-dependent, env-observing)."""
        return cls()

    @classmethod
    def stderr(cls) -> Color:
        """Create a Color that checks ``sys.stderr`` instead of ``sys.stdout``.

        Useful for rendering error messages, which typically go to stderr.
        Respects ``NO_COLOR`` / ``FORCE_COLOR`` normally.
        """
        c = cls()
        c._check_stderr = True
        return c

    @property
    def is_terminal(self) -> bool:
        """True if the configured stream is connected to a terminal.

        By default checks ``sys.stdout``; ``Color.stderr()`` checks
        ``sys.stderr`` instead.
        """
        import sys

        stream = sys.stderr if self._check_stderr else sys.stdout
        return stream.isatty()

    @staticmethod
    def terminal_size() -> os.terminal_size:
        """Return ``(columns, lines)`` of the terminal, falling back to 80x24."""
        import shutil

        return shutil.get_terminal_size()

    @property
    def supports_color(self) -> bool:
        """True if the terminal is capable of ANSI color (independent of policy).

        Checks TTY, ``TERM`` (not dumb/emacs), and terminal width.
        Does NOT check ``NO_COLOR``/``FORCE_COLOR`` — see ``enabled`` for that.
        """
        import os
        import shutil
        import sys

        if not sys.stdout.isatty():
            return False
        if os.environ.get("TERM", "") in {"dumb", "emacs"}:
            return False
        try:
            size = shutil.get_terminal_size()
            if size.columns < 1:
                return False
        except (ValueError, OSError):
            return False
        return True

    @property
    def enabled(self) -> bool:
        """Whether color should be emitted, evaluating policy in priority order:

        1. Explicit ``enable()`` override.
        2. ``NO_COLOR`` env var → disabled.
        3. ``FORCE_COLOR`` env var → enabled.
        4. ``sys.stdout.isatty()``.
        """
        import os

        if self._force_enable is not None:
            return self._force_enable
        if os.environ.get("NO_COLOR") is not None:
            return False
        if os.environ.get("FORCE_COLOR") is not None:
            return True
        return self.is_terminal


DEFAULT_COLOR: Color = Color.default()


class Style(str):
    """A composable ANSI style builder.

    ``Style`` stores a text *value* plus formatting attributes (foreground
    color, background color, bold, dim, italic, …).  All modifier methods
    return a **copy** so styles are immutable and chainable:

    .. code-block:: python

        s = Style("hello", bold=True, fg=2)
        str(s)  # → "\\033[1;32mhello\\033[0m"

    When ``color.enabled`` is False, ``str()`` and ``apply()`` return the
    plain text unchanged — no ANSI codes leak into pipes, logs, or CI output.

    Most modifier methods accept a ``None`` or ``-1`` fg/bg to unset
    a previously assigned color, and ``RGB`` for 24-bit color.
    """

    __slots__ = (
        "_bg",
        "_blink",
        "_bold",
        "_color",
        "_dim",
        "_fg",
        "_fmt",
        "_hidden",
        "_inverse",
        "_italic",
        "_strikethrough",
        "_underline",
    )

    def __json__(self, seen: set[int] | None = None) -> Any:
        return repr(self)

    def markup(self, text: str):
        from .markup import markup

        return markup(text, color=self.color)

    @property
    def value(self) -> str:
        return super().__str__()

    @property
    def color(self) -> Color:
        return self._color

    @classmethod
    def tty_escape(cls, text: str) -> str:
        return tty_escape(text)

    @classmethod
    def tty_unescape(cls, text: str) -> str:
        return tty_unescape(text)

    @classmethod
    def is_style_method(cls, name: str) -> bool:
        meth: Callable[[], Self] | None = None
        if (meth := getattr(cls, name, None)) is None or not callable(meth):
            return False
        sig = inspect.signature(meth)
        if str(sig.return_annotation) not in {"Style", "Self"}:
            return False
        with contextlib.suppress(TypeError):
            sig.bind(None)
            return True
        return False

    def get_style_method(self, name: str) -> Callable[[], Self] | None:
        if not self.is_style_method(name):
            return None
        return getattr(self, name, None)

    @classmethod
    def list_style_methods(cls) -> list[str]:
        return [name for name in dir(cls) if cls.is_style_method(name)]

    def __new__(cls, *args, **_kwargs):
        return super().__new__(cls, *args)

    def __str__(self) -> str:
        return self.apply(self.value)

    def __format__(self, format_spec: str) -> str:
        return self.apply(str(self), fmt=format_spec)

    def __repr__(self) -> str:
        text = self.value
        if self._fmt is not None:
            # NOTE Meant for Style.parse() not for f-strings
            text = f"f{{{self.value}:{self._fmt}}}"
        text = self.apply_style(text, force=True)
        return tty_escape(repr(text)[1:-1])

    def __len__(self) -> int:
        return visual_len(str(self))

    def __hash__(self) -> int:
        return hash(self.__dict__)

    def __eq__(self, other: object) -> bool:
        if not hasattr(other, "__dict__"):
            return False
        return self.__dict__ == other.__dict__

    def __init__(
        self,
        value: str = "",
        /,
        *,
        fmt: str | None = None,
        fg: int | RGB | None = -1,
        bg: int | RGB | None = -1,
        bold: bool = False,
        dim: bool = False,
        italic: bool = False,
        underline: bool = False,
        blink: bool = False,
        inverse: bool = False,
        hidden: bool = False,
        strikethrough: bool = False,
        color: Color = DEFAULT_COLOR,
        **_kwargs,
    ):
        super().__init__()
        _ = value

        self._color = color

        self._fg: int | RGB = -1
        self._bg: int | RGB = -1

        self._set_fg(fg)
        self._set_bg(bg)

        self._fmt = fmt
        self._bold = bold
        self._dim = dim
        self._italic = italic
        self._underline = underline
        self._blink = blink
        self._inverse = inverse
        self._hidden = hidden
        self._strikethrough = strikethrough

    def _kwattrs(self) -> dict[str, Any]:
        return {k.lstrip('_'): getattr(self, k, None) for k in type(self).__slots__}

    def __call__(self, value: str, /, *, fmt: str | None = None) -> Self:
        kw = self._kwattrs()
        kw.pop('value', None)
        if fmt is not None:
            kw['fmt'] = fmt
        return type(self)(value, **kw)

    @property
    def enabled(self) -> bool:
        """Whether this style emits ANSI codes, delegated to the bound ``Color``."""
        return self._color.enabled

    def _set_fg(self, value: int | RGB | None) -> None:
        if isinstance(value, RGB):
            self._fg = value
        elif value is None or value < 0:
            self._fg = -1
        else:
            self._fg = max(0, min(value, 255))

    def _set_bg(self, value: int | RGB | None) -> None:
        if isinstance(value, RGB):
            self._bg = value
        elif value is None or value < 0:
            self._bg = -1
        else:
            self._bg = max(0, min(value, 255))

    def fg_name(self, name: str) -> Self:
        """Set foreground to a named ANSI color (e.g. ``"red"``, ``"bright_green"``).

        The color name is matched case-insensitively, with spaces removed.
        This loads the color map lazily on first call.
        """
        from .colormap import color

        new = copy(self)
        new._set_fg(color(name))
        return new

    def bg_name(self, name: str) -> Self:
        """Set background to a named ANSI color (e.g. ``"navy"``, ``"teal"``).

        Like ``fg_name``, the color map loads lazily.
        """
        from .colormap import color

        new = copy(self)
        new._set_bg(color(name))
        return new

    def fg_css(self, name: str) -> Self:
        """Set foreground to a CSS named color (e.g. ``"rebeccapurple"``).

        Uses the 148 CSS named colors.  Loads lazily on first call.
        """
        from .csscolormap import css_color

        new = copy(self)
        rgb = css_color(name)
        if rgb is not None:
            new._set_fg(rgb)
        return new

    def bg_css(self, name: str) -> Self:
        """Set background to a CSS named color (e.g. ``"lightseagreen"``)."""
        from .csscolormap import css_color

        new = copy(self)
        rgb = css_color(name)
        if rgb is not None:
            new._set_bg(rgb)
        return new

    def fg(self, value: int | RGB | None) -> Self:
        """Set foreground to an ANSI 256-color index (0-255) or RGB.

        Pass ``None`` to reset the foreground color.
        """
        new = copy(self)
        new._set_fg(value)
        return new

    def bg(self, value: int | RGB | None) -> Self:
        """Set background to an ANSI 256-color index (0-255) or RGB.

        Pass ``None`` to reset the background color.
        """
        new = copy(self)
        new._set_bg(value)
        return new

    def fg_rgb(self, r: int, g: int, b: int) -> Self:
        """Set foreground to a 24-bit RGB color."""
        new = copy(self)
        new._set_fg(rgb(r, g, b))
        return new

    def bg_rgb(self, r: int, g: int, b: int) -> Self:
        """Set background to a 24-bit RGB color."""
        new = copy(self)
        new._set_bg(rgb(r, g, b))
        return new

    # ANSI standard colors (0-7)

    def black(self) -> Self:
        return self.fg(0)

    def black_bg(self) -> Self:
        return self.bg(0)

    def red(self) -> Self:
        return self.fg(1)

    def red_bg(self) -> Self:
        return self.bg(1)

    def green(self) -> Self:
        return self.fg(2)

    def green_bg(self) -> Self:
        return self.bg(2)

    def yellow(self) -> Self:
        return self.fg(3)

    def yellow_bg(self) -> Self:
        return self.bg(3)

    def blue(self) -> Self:
        return self.fg(4)

    def blue_bg(self) -> Self:
        return self.bg(4)

    def purple(self) -> Self:
        return self.fg(5)

    def purple_bg(self) -> Self:
        return self.bg(5)

    def magenta(self) -> Self:
        return self.purple()

    def cyan(self) -> Self:
        return self.fg(6)

    def cyan_bg(self) -> Self:
        return self.bg(6)

    def white(self) -> Self:
        return self.fg(7)

    def white_bg(self) -> Self:
        return self.bg(7)

    # ANSI bright colors (8-15)

    def bright_black(self) -> Self:
        return self.fg(8)

    def bright_black_bg(self) -> Self:
        return self.bg(8)

    def bright_red(self) -> Self:
        return self.fg(9)

    def bright_red_bg(self) -> Self:
        return self.bg(9)

    def bright_green(self) -> Self:
        return self.fg(10)

    def bright_green_bg(self) -> Self:
        return self.bg(10)

    def bright_yellow(self) -> Self:
        return self.fg(11)

    def bright_yellow_bg(self) -> Self:
        return self.bg(11)

    def bright_blue(self) -> Self:
        return self.fg(12)

    def bright_blue_bg(self) -> Self:
        return self.bg(12)

    def bright_purple(self) -> Self:
        return self.fg(13)

    def bright_purple_bg(self) -> Self:
        return self.bg(13)

    def bright_magenta(self) -> Self:
        return self.bright_purple()

    def bright_cyan(self) -> Self:
        return self.fg(14)

    def bright_cyan_bg(self) -> Self:
        return self.bg(14)

    def bright_white(self) -> Self:
        return self.fg(15)

    def bright_white_bg(self) -> Self:
        return self.bg(15)

    def bold(self) -> Self:
        """Add bold (SGR 1)."""
        new = copy(self)
        new._bold = True
        return new

    def bright(self) -> Self:
        """Add bright (SGR 1)."""
        return self.bold()

    def dim(self) -> Self:
        """Add dim/faint (SGR 2)."""
        new = copy(self)
        new._dim = True
        return new

    def italic(self) -> Self:
        """Add italic (SGR 3)."""
        new = copy(self)
        new._italic = True
        return new

    def underline(self) -> Self:
        """Add underline (SGR 4)."""
        new = copy(self)
        new._underline = True
        return new

    def blink(self) -> Self:
        """Add slow blink (SGR 5)."""
        new = copy(self)
        new._blink = True
        return new

    def inverse(self) -> Self:
        """Swap foreground and background colors (SGR 7)."""
        new = copy(self)
        new._inverse = True
        return new

    def hidden(self) -> Self:
        """Hide text / make invisible (SGR 8)."""
        new = copy(self)
        new._hidden = True
        return new

    def strikethrough(self) -> Self:
        """Add strikethrough / crossed-out (SGR 9)."""
        new = copy(self)
        new._strikethrough = True
        return new

    def fmt(self, fmt_spec: str) -> Self:
        """Set a format specification (a style will still create a copy)."""
        new = copy(self)
        new._fmt = fmt_spec
        return new

    def apply(self, text: str, fmt: str | None = None) -> str:
        """Wrap *text* in ANSI escape codes according to this style.

        If ``self.enabled`` is False, returns *text* unchanged.
        If *text* is empty, returns ``""``.
        When no formatting attributes are set, returns the stored ``self.value``
        (the behaviour that powers ``__str__``).
        """
        if not text:
            return ""
        if not fmt:
            fmt = self._fmt
        text = format(text, fmt) if fmt else text
        return self.apply_style(text)

    def apply_style(self, text: str, force: bool = False) -> str:
        if not text:
            return ""
        if not (self.enabled or force):
            return text
        codes: list[str] = []
        if self._bold:
            codes.append('1')
        if self._dim:
            codes.append('2')
        if self._italic:
            codes.append('3')
        if self._underline:
            codes.append('4')
        if self._blink:
            codes.append('5')
        if self._inverse:
            codes.append('7')
        if self._hidden:
            codes.append('8')
        if self._strikethrough:
            codes.append('9')
        if isinstance(self._fg, RGB):
            codes.append(f'38;2;{self._fg.r};{self._fg.g};{self._fg.b}')
        elif self._fg != -1:
            c = self._fg
            if c < 8:
                codes.append(str(30 + c))
            elif c < 16:
                codes.append(str(90 + c - 8))
            else:
                codes.append(f'38;5;{c}')
        if isinstance(self._bg, RGB):
            codes.append(f'48;2;{self._bg.r};{self._bg.g};{self._bg.b}')
        elif self._bg != -1:
            c = self._bg
            if c < 8:
                codes.append(str(40 + c))
            elif c < 16:
                codes.append(str(100 + c - 8))
            else:
                codes.append(f'48;5;{c}')
        if not codes:
            return text
        return f"\033[{';'.join(codes)}m{text}\033[0m"

    @classmethod
    def from_raw(cls, text: str) -> Self:
        text = tty_unescape(text)

        sgr = SGR_RE.search(text)
        if not sgr:
            return cls.parse_fmt(text)

        text_content = ANSI_RE.sub("", text)
        params_str = sgr.group(1)

        fg: int | RGB = -1
        bg: int | RGB = -1
        bold = dim = italic = underline = blink = inverse = hidden = strikethrough = (
            False
        )

        if params_str:
            params = [int(p) for p in params_str.split(";")]
            i = 0
            while i < len(params):
                p = params[i]
                if p == 0:
                    pass
                elif p == 1:
                    bold = True
                elif p == 2:
                    dim = True
                elif p == 3:
                    italic = True
                elif p == 4:
                    underline = True
                elif p == 5:
                    blink = True
                elif p == 7:
                    inverse = True
                elif p == 8:
                    hidden = True
                elif p == 9:
                    strikethrough = True
                elif 30 <= p <= 37:
                    fg = p - 30
                elif 40 <= p <= 47:
                    bg = p - 40
                elif p == 38:
                    if i + 1 < len(params):
                        if params[i + 1] == 5 and i + 2 < len(params):
                            fg = params[i + 2]
                            i += 2
                        elif params[i + 1] == 2 and i + 4 < len(params):
                            fg = RGB(params[i + 2], params[i + 3], params[i + 4])
                            i += 4
                elif p == 48:
                    if i + 1 < len(params):
                        if params[i + 1] == 5 and i + 2 < len(params):
                            bg = params[i + 2]
                            i += 2
                        elif params[i + 1] == 2 and i + 4 < len(params):
                            bg = RGB(params[i + 2], params[i + 3], params[i + 4])
                            i += 4
                elif 90 <= p <= 97:
                    fg = p - 82
                elif 100 <= p <= 107:
                    bg = p - 92
                i += 1

        kwargs: dict[str, Any] = {}
        if fg != -1:
            kwargs["fg"] = fg
        if bg != -1:
            kwargs["bg"] = bg
        if bold:
            kwargs["bold"] = True
        if dim:
            kwargs["dim"] = True
        if italic:
            kwargs["italic"] = True
        if underline:
            kwargs["underline"] = True
        if blink:
            kwargs["blink"] = True
        if inverse:
            kwargs["inverse"] = True
        if hidden:
            kwargs["hidden"] = True
        if strikethrough:
            kwargs["strikethrough"] = True

        return cls.parse_fmt(text_content, **kwargs)

    @classmethod
    def parse_fmt(cls, text: str, **kwargs) -> Self:
        fmt = None
        if m := re.match(r"f{(.*?):(.*)}", text):
            text, fmt = m.group(1, 2)
        return cls(text, color=Color.default(), fmt=fmt, **kwargs)


def named_color(name: str) -> int | None:
    """Look up a named ANSI color code (0-255) by name.

    The color name is matched case-insensitively, with spaces removed.
    Returns the ANSI 256-color index or ``None`` if unknown.
    The color-map module is loaded lazily.
    """
    from .colormap import color

    return color(name)


def css_color(name: str) -> RGB | None:
    """Look up a CSS named color as an ``RGB`` tuple.

    Accepts standard CSS color names like ``"rebeccapurple"``,
    ``"mediumseagreen"``, etc.  Returns ``None`` for unknown names.
    The CSS color map is loaded lazily.
    """
    from .csscolormap import css_color

    return css_color(name)


def fmt(value: str, fmt: str) -> str:
    """Format *value* according to the Python format specification *fmt*.

    Wraps ``f'{value:{fmt}}'`` with exception handling so invalid
    format specs degrade gracefully to the original value.

    Example:
        fmt("hello", ">10")  # "     hello"
        fmt("hello", ".2")   # "he"
    """
    if not fmt:
        return value
    try:
        return f'{value:{fmt}}'
    except (ValueError, TypeError):
        return value



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys

from . import config as parserconfig, input as tokenizing
from ._grammar import grammar, grammar_path
from ._version import __toolname__, __version__, version, version_info
from .api import (
    compile,
    compile_to_parser,
    gencode,
    genmodel,
    parse,
    to_grammar_json,
    to_python_model,
    to_python_sourcecode,
)  # pylint: disable=W0622
from .contexts import ast as ast
from .contexts.decorator import isname, leftrec, name, nomemo, rule, tatsumasu
from .input import buffer as buffer, buffer as buffering, textlines as textlines
from .objectmodel import (
    NodeDataclassParams,
    NodeDataclassParams as TatSuDataclassParams,
    builder as builder,
    nodedataclass,
    nodedataclass as dataclass,
    nodedataclass as tatsudataclass,
)
from .tool.cli import tatsu_main


# HACK!
# NOTE: this is for backwrds compatibility with legacy generated parsers
sys.modules.update(  # noqa: RUF067
    {
        'tatsu.ast': ast,
        'tatsu.builder': builder,
        'tatsu.buffering': buffering,
        'tatsu.parserconfig': parserconfig,
        'tatsu.tokenizing': tokenizing,
        'tatsu.tokenizing.buffer': buffer,
        'tatsu.tokenizing.textlines': textlines,
    }
)


__all__ = [
    '__toolname__',
    '__version__',
    'compile_to_parser',
    'builder',
    'compile',
    'gencode',
    'genmodel',
    'parse',
    'tatsu_main',
    'to_grammar_json',
    'to_python_model',
    'to_python_sourcecode',
    'version',
    'version_info',
    'isname',
    'NodeDataclassParams',
    'TatSuDataclassParams',
    'dataclass',
    'nodedataclass',
    'tatsudataclass',
    'name',
    'leftrec',
    'nomemo',
    'rule',
    'tatsumasu',
    'grammar_path',
    'grammar',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from dataclasses import field
from functools import cached_property
from typing import Any

from ..contexts import Ctx
from ..exceptions import FailedParse
from ..objectmodel import nodedataclass
from .base import PEP8_LLEN, Box, Model
from .math import ffset


@nodedataclass
class Option(Box):
    # NOTE parsing of options is handled by Choice._parse()
    # def _parse(self, ctx: Ctx) -> Any:

    def optimized(self) -> Model:
        # NOTE hocus pocus! now you see me...
        return self.exp.optimized()  # ... now you don't


@nodedataclass
class Choice(Model):
    options: list[Option] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.options = self.options or self.ast
        assert isinstance(self.options, list), repr(self.options)
        assert self.options, repr(self.options)

    def _parse(self, ctx: Ctx) -> Any:
        # ctx.expecting(*self.expecting)
        for o in self.options:
            ctx.states.push()
            o._add_defined(ctx)
            exp = o.exp if isinstance(o, Option) else o
            try:
                value = exp._parse(ctx)
                ctx.states.merge()
                return value
            except FailedParse:
                if ctx.states.undo().cutseen:
                    raise
        raise ctx.newexcept(self.expectingstr)

    @cached_property
    def defines_single(self) -> list[str]:
        return list(set().union(*(o.defines_single for o in self.options)))

    @cached_property
    def defines_list(self) -> list[str]:
        return list(set().union(*(o.defines_list for o in self.options)))

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set().union(*[o.missing_rules(rulenames) for o in self.options])

    def _used_rule_names(self):
        return set().union(*[o._used_rule_names() for o in self.options])

    def _first(self, k, f) -> ffset:
        result = set()
        for o in self.options:
            result |= o._first(k, f)
        self._firstset = result
        return result

    def _follow(self, k, fl, a):
        for o in self.options:
            o._follow(k, fl, a)
        return a

    def nodecount(self) -> int:
        return 1 + sum(o.nodecount() for o in self.options)

    def _pretty(self, lean=False):
        options = [str(o._pretty(lean=lean)) for o in self.options]

        multi = options and any(len(o.splitlines()) > 1 for o in options)
        single = ' | '.join(o for o in options)
        multi = multi or len(single) > PEP8_LLEN * 0.6  # WARNING: magic number !

        if multi:
            return '| ' + '\n| '.join(o for o in options)
        else:
            return single

    @cached_property
    def _nullable(self) -> bool:
        return any(o._nullable for o in self.options)

    def optimized(self) -> Model | Choice:
        opt = [o.optimized() for o in self.options]
        for o in opt:
            assert isinstance(o, Model)
        if len(opt) == 1:
            return opt[0]
        return Choice(options=opt)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import unittest

import pytest  # pyright: ignore [reportMissingImports]

from tatsu import compile, parse
from tatsu.boot import TatSuBuffer
from tatsu.exceptions import FailedParse
from tatsu.ngcodegen import pythongen
from tatsu.util import trim


class SyntaxTests(unittest.TestCase):
    def test_update_ast(self):
        grammar = """
            foo = name:"1" [ name: bar ] ;
            bar = { "2" } * ;
        """
        m = compile(grammar, 'Keywords')
        ast = m.parse('1 2')
        self.assertEqual(['1', ['2']], ast.name)

        grammar = """
            start = items: { item } * $ ;
            item = @:{ subitem } * "0" ;
            subitem = /1+/ ;
        """
        m = compile(grammar, 'Update')
        ast = m.parse('1101110100', nameguard=False)
        self.assertEqual([['11'], ['111'], ['1'], []], ast.items_)

    def test_include_and_override(self):
        gr = 'included_grammar'
        included_grammar = "plu = 'aaaa';"

        overridden = "%s@override\nplu = 'plu';"
        inclusion = f'#include :: {gr}.tatsu\n'

        including_grammar = overridden % inclusion
        whole_grammar = overridden % included_grammar

        class FakeIncludesBuffer(TatSuBuffer):
            def get_include(self, source, filename):
                return included_grammar, source + '/' + filename

        compile(FakeIncludesBuffer(whole_grammar), 'test')
        compile(FakeIncludesBuffer(including_grammar), 'test')

    def test_ast_assignment(self):
        grammar = """
            n  = @: {"a"}* $ ;
            f  = @+: {"a"}* $ ;
            nn = @: {"a"}*  @: {"b"}* $ ;
            nf = @: {"a"}*  @+: {"b"}* $ ;
            fn = @+: {"a"}* @: {"b"}* $ ;
            ff = @+: {"a"}* @+: {"b"}* $ ;
        """

        model = compile(grammar, 'test')

        def p(input, rule):
            return model.parse(input, start=rule, whitespace='')

        e = self.assertEqual

        e([], p('', 'n'))
        e(['a'], p('a', 'n'))
        e(['a', 'a'], p('aa', 'n'))

        e([[]], p('', 'f'))
        e([['a']], p('a', 'f'))
        e([['a', 'a']], p('aa', 'f'))

        for r in ('nn', 'nf', 'fn', 'ff'):
            e([[], []], p('', r), r)
            e([['a'], []], p('a', r), r)
            e([[], ['b']], p('b', r), r)
            e([['a', 'a'], []], p('aa', r), r)
            e([[], ['b', 'b']], p('bb', r), r)
            e([['a', 'a'], ['b']], p('aab', r), r)

    def test_optional_closure(self):
        grammar = 'start = foo+:"x" foo:{"y"}* {foo:"z"}* ;'
        model = compile(grammar, 'test')
        ast = model.parse('xyyzz', nameguard=False)
        self.assertEqual(['x', ['y', 'y'], 'z', 'z'], ast.foo)

        grammar = 'start = foo+:"x" [foo+:{"y"}*] {foo:"z"}* ;'
        model = compile(grammar, 'test')
        ast = model.parse('xyyzz', nameguard=False)
        self.assertEqual(['x', ['y', 'y'], 'z', 'z'], ast.foo)

        grammar = 'start = foo+:"x" foo:[{"y"}*] {foo:"z"}* ;'
        model = compile(grammar, 'test')
        ast = model.parse('xyyzz', nameguard=False)
        self.assertEqual(['x', ['y', 'y'], 'z', 'z'], ast.foo)

        grammar = 'start = foo+:"x" [foo:{"y"}*] {foo:"z"}* ;'
        model = compile(grammar, 'test')
        ast = model.parse('xyyzz', nameguard=False)
        self.assertEqual(['x', ['y', 'y'], 'z', 'z'], ast.foo)

    def test_optional_sequence(self):
        grammar = """
            start = '1' ['2' '3'] '4' $ ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('1234', nameguard=False)
        self.assertEqual(['1', '2', '3', '4'], ast)

        grammar = """
            start = '1' foo:['2' '3'] '4' $ ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('1234', nameguard=False)
        self.assertEqual(['2', '3'], ast.foo)

    def test_group_ast(self):
        grammar = """
            start = '1' ('2' '3') '4' $ ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('1234', nameguard=False)
        self.assertEqual(['1', '2', '3', '4'], ast)

    def test_partial_options(self):
        grammar = """
            start
                =
                [a]
                [
                    'A' 'A'
                |
                    'A' 'B'
                ]
                $
                ;
            a
                =
                'A' !('A'|'B')
                ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('AB', nameguard=False)
        self.assertEqual(['A', 'B'], ast)

    def test_partial_choice(self):
        grammar = """
            start
                =
                o:[o]
                x:'A'
                $
                ;
            o
                =
                'A' a:'A'
                |
                'A' b:'B'
                ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('A', nameguard=False)
        self.assertEqual({'x': 'A', 'o': None}, ast)

    def test_new_override(self):
        grammar = """
            start
                =
                @:'a' {@:'b'}
                $
                ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('abb', nameguard=False)
        self.assertEqual(['a', 'b', 'b'], ast)

    def test_list_override(self):
        grammar = """
            start
                =
                @+:'a' {@:'b'}
                $
                ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('a', nameguard=False)
        self.assertEqual(['a'], ast)

        grammar = """
            start
                =
                @:'a' {@:'b'}
                $
                ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('a', nameguard=False)
        self.assertEqual('a', ast)

    def test_based_rule(self):
        grammar = """\
            start: b $

            a: ='a'

            b < a: {='b'}
            """
        model = compile(grammar, 'test')
        ast = model.parse('abb', nameguard=False)
        self.assertEqual(['a', 'b', 'b'], ast)
        self.assertEqual(trim(grammar), trim(model.pretty()))

    def test_rule_include(self):
        grammar = """
            start = b $;

            a = @:'a' ;
            b = >a {@:'b'} ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('abb', nameguard=False)
        self.assertEqual(['a', 'b', 'b'], ast)

    def test_48_rule_override(self):
        grammar = """
            start = ab $;

            ab = 'xyz' ;

            @override
            ab = @:'a' {@:'b'} ;
        """
        model = compile(grammar, 'test')
        ast = model.parse('abb', nameguard=False)
        self.assertEqual(['a', 'b', 'b'], ast)

    def test_failed_ref(self):
        grammar = r"""
            final = object;
            type = /[^\s=()]+/;
            object = '('type')' '{' @:{pair} {',' @:{pair}}* [','] '}';
            pair = key '=' value;
            list = '('type')' '[' @:{object} {',' @:{object}}* [','] ']';
            key = /[^\s=]+/;
            value = @:(string|list|object|unset|boolean|number|null) [','];
            null = '('type')' @:{ 'null' };
            boolean = /(true|false)/;
            unset = '<unset>';
            string = '"' @:/[^"]*/ '"';
            number = /-?[0-9]+/;
        """

        model = compile(grammar, 'final')
        pythongen(model)
        model.parse('(sometype){boolean = true}')

    def test_empty_match_token(self):
        grammar = """
            table = { row }+ ;
            row = (cell1:cell "|" cell2:cell) "\n";
            cell = /[a-z]+/ ;
        """
        try:
            compile(grammar, 'model')
            self.fail('allowed empty token')
        except FailedParse:
            pass

    def test_empty_closure(self):
        grammar = """
            start = {'x'}+ {} 'y'$;
        """
        model = compile(grammar, 'test')
        pythongen(model)
        ast = model.parse('xxxy', nameguard=False)
        self.assertEqual([['x', 'x', 'x'], [], 'y'], ast)

    def test_parseinfo(self):
        grammar = """
            start = head:{'x'}+ {} tail:'y'$;
        """
        model = compile(grammar, 'test')
        ast = model.parse('xxxy', nameguard=False, parseinfo=True)
        self.assertIsNotNone(ast)
        self.assertIsNotNone(ast.head)
        self.assertIsNotNone(ast.tail)
        self.assertIsNotNone(ast.parseinfo)

    def test_raw_string(self):
        grammar = r"""
            start
                =
                r'am\nraw'
                ;
        """
        pretty = r"""
            start: 'am\\nraw'

        """
        model = compile(grammar, 'start')
        print(model.pretty())
        self.assertEqual(trim(pretty), model.pretty())

    def test_any(self):
        grammar = """
            start = /./ 'xx' /./ /./ 'yy' $;
        """
        model = compile(grammar, 'start')
        ast = model.parse('1xx 2 yy')
        self.assertEqual(['1', 'xx', ' ', '2', 'yy'], ast)

    def test_constant(self):
        grammar = """
            start = ()
                _0:`0` _1:`+1` _n123:`-123`
                _xF:`0xF`
                _string:`string`
                _string_space:`'string space'`
                _true:`True` _false:`False`
                $;
        """

        model = compile(grammar)
        ast = model.parse('')

        self.assertEqual(ast._0, 0)
        self.assertEqual(ast._1, 1)
        self.assertEqual(ast._n123, -123)
        self.assertEqual(ast._xF, 0xF)
        self.assertEqual(ast._string, 'string')
        self.assertEqual(ast._string_space, 'string space')
        self.assertEqual(ast._true, True)
        self.assertEqual(ast._false, False)


def test_parse_hash():
    grammar = r"""
        @@comments :: /@@@@@@@/
        @@eol_comments :: /@@@@@@@/

        start = '#' ;
    """

    parser = compile(grammar, eol_comments='')
    parser.parse('#', trace=True)


def test_parse_void():
    grammar = r"""
        start = () $ ;
    """

    ast = parse(grammar, '')
    print(ast)
    assert ast is None


def test_no_default_comments():
    grammar = """
        start = 'a' $;
    """

    text = """
        # no comments are valid
        a
    """
    with pytest.raises(FailedParse):
        parse(grammar, text)


import re


@pytest.mark.parametrize(
    "comment,option",
    [
        pytest.param(
            "# This comment should be stripped",
            {
                "eol_comments_re": r"(?m)#[^\n]*$",
                "comments": re.compile(r"(?sm)[(][*](?:.|\n)*?[*][)]"),
            },
            id="eol_comments override",
        ),
        pytest.param(
            "(* This comment should be stripped *)",
            {
                "eol_comments": re.compile(r"(?m)#[^\n]*$"),
                "comments_re": r"(?sm)[(][*](?:.|\n)*?[*][)]",
            },
            id="comments override",
        ),
    ],
)
def test_deprecated_comments_override_failures(comment, option):
    """
    # TODO: remove this test after {eol_}comments_re are no longer supported
    """
    grammar = """
        @@comments :: /@@@@@@/
        @@eol_comments :: /@@@@@@/

        start = 'a' $;
    """

    text = f"""
        {comment}
        a
    """
    with pytest.warns(
        DeprecationWarning,
        match=r'ParserConfig\..*?comments_re.*?is deprecated',
    ):
        parse(grammar, text, **option)


def test_non_capturing_group_exclusion():
    # A grammar where 'header' is captured, but 'delimiter' is skipped
    grammar = """
    start: header (?: delimiter ) body

    header: /[A-Z]+/

    delimiter: /[:,-]+/

    body: /[a-z]+/
    """
    parser = compile(grammar)

    input_str = "INFO---data"
    ast = parser.parse(input_str)

    # Assertions
    assert ast[0] == "INFO"
    assert ast[1] == "data"
    # The AST should only have 2 elements because the delimiter was (?: ...)
    assert len(ast) == 2

    # Verify the delimiter is not present in the values
    assert "---" not in ast


def test_non_capturing_group_failure():
    # Verify that even if not captured, the expression must still match
    grammar = r"start = (?: 'FIX' ) value ; value = /\d+/"
    parser = compile(grammar)

    # This should fail because 'BUG' does not match 'FIX'
    with pytest.raises(FailedParse):
        parser.parse("BUG123")


def test_nested_non_capturing_groups():
    grammar = """
    start: (?: '(' (?: inner ) ')' )

    inner: 'content'
    """
    parser = compile(grammar)

    ast = parser.parse("(content)")

    # Since the outer group is non-capturing, the AST for 'start'
    # should be None or empty depending on the surrounding context.
    assert ast is None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import weakref
from collections import defaultdict
from collections.abc import Iterable, Mapping
from copy import copy
from dataclasses import field
from functools import cached_property
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Self

from ..config import ParserConfig
from ..contexts import AST, CanParse, Ctx, Func, ParseContext, RuleInfo
from ..exceptions import GrammarError
from ..input import Text
from ..objectmodel import ModelBuilderSemantics, Node, nodedataclass
from ..util import indent, trim, typename
from ..util.strtools import slicetowidth
from ..util.undefined import UndefinedType
from .math import ffset, kdot


PEP8_LLEN = 72

_model_classes: list[type[Model]] = []


def model_classes() -> list[type[Model]]:
    return _model_classes


@nodedataclass
class Model(Node, CanParse):
    _lookahead: ffset = field(init=False, default_factory=set)
    _firstset: ffset = field(init=False, default_factory=set)
    _follow_set: ffset = field(init=False, default_factory=set)
    _grammar_ref: weakref.ref[Model] | None = field(init=False, default=None)

    def __init_subclass__(cls: type[Model], **kwargs):
        super().__init_subclass__(**kwargs)
        _model_classes.append(cls)

    @staticmethod
    def model_classes() -> list[type[Model]]:
        return _model_classes

    def follow_ref(self) -> Model:
        return self

    @cached_property
    def defines_single(self) -> list[str]:
        return []

    @cached_property
    def defines_list(self) -> list[str]:
        return []

    @property
    def grammar(self) -> Grammar:
        if self._grammar_ref is None:
            raise RuntimeError(f'{typename(self)} not fully initialized')
        grammar = self._grammar_ref()
        if not isinstance(grammar, Grammar):
            raise TypeError(f'{typename(self)} incorrectly initialized {grammar}')
        return grammar

    def parse(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> Any:
        grammar = self.grammar
        assert isinstance(grammar, Grammar)
        return grammar.parse(
            text,
            start=start,
            config=config,
            asmodel=asmodel,
            **settings,
        )

    def _parse(self, ctx: Ctx) -> Any:
        _ = ctx
        return ()

    def link(self, grammar: Grammar):
        assert isinstance(grammar, Grammar)

        self._grammar_ref = weakref.ref(grammar)
        for child in self.children():
            assert isinstance(child, Model)
            child.link(grammar)

    def _add_defined(self, ctx: Ctx):
        keys_list = self.defines_list
        keys_single = [d for d in self.defines_single if d not in keys_list]
        ctx.define(keys_single, keys_list)

    def lookahead(self, k: int = 1) -> ffset:
        if not getattr(self, '_lookahead', None):
            self._lookahead = kdot(self.firstset(k), self.followset(k), k)
        return self._lookahead

    @cached_property
    def lookaheadlist(self) -> list[Any]:
        return sorted(fl[0] for fl in self.lookahead(1) if fl)

    @cached_property
    def expecting(self) -> list[str]:
        return sorted(slicetowidth(str(la), 40) for la in self.lookaheadlist)

    @cached_property
    def expectingstr(self) -> str:
        return f'expecting one of {' '.join(repr(s) for s in self.expecting)}'

    def firstset(self, k: int = 1) -> ffset:
        if not getattr(self, '_firstset', None):
            self._firstset = self._first(k, defaultdict(set))
        return self._firstset

    def followset(self, _k: int = 1) -> ffset:
        if not getattr(self, '_follow_set', None):
            self._follow_set = set()
        return self._follow_set

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        assert rulenames
        return set()

    def _used_rule_names(self):
        return set()

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:  # pyright: ignore[reportUnusedParameter]
        _ = k
        _ = f
        return set()

    def _follow(self, k, fl, a):
        _ = k
        _ = fl
        return a

    def is_nullable(self) -> bool:
        return self._nullable

    @cached_property
    def _nullable(self) -> bool:
        return False

    def nodecount(self) -> int:
        return 1

    def pretty(self, lean: bool = False) -> str:
        return self._pretty(lean=lean)

    def pretty_lean(self):
        return self._pretty(lean=True)

    def _pretty(self, lean=False):
        _ = lean
        return f'{typename(self)}: {id(self)}'

    def railroads(self) -> str:
        from .. import railroads

        return railroads.text(self)

    def optimized(self) -> Model:
        return copy(self)


@nodedataclass
class Leaf(Model):
    pass


@nodedataclass
class NIL(Leaf):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.fail() or ()

    def _pretty(self, lean=False):
        _ = lean
        return typename(self)

    @cached_property
    def _nullable(self) -> bool:
        return True


@nodedataclass
class Void(Leaf):
    def __post_init__(self):
        super().__post_init__()
        self.ast = None

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.void()

    def _pretty(self, lean=False):
        _ = lean
        return '()'

    @cached_property
    def _nullable(self) -> bool:
        return True


@nodedataclass
class Box(Model):
    name: str | None = field(init=False, default=None)
    exp: Model = field(default_factory=NIL)

    def __post_init__(self):
        noexp = not self.exp or isinstance(self.exp, NIL)
        if noexp and not isinstance(self.ast, AST):
            self.exp = self.ast
        super().__post_init__()
        if isinstance(self.ast, AST):
            assert self.exp == self.ast.exp
        # assert not isinstance(self.exp, NIL), self.ast
        self.exp = self.exp if self.exp is not None else NIL()
        assert self.exp is not None, f'{typename(self)}({self.exp})'
        assert self.exp, repr(self)

    def _parse(self, ctx: Ctx) -> Any:
        return self.exp._parse(ctx)

    @cached_property
    def defines_single(self) -> list[str]:
        return self.exp.defines_single

    @cached_property
    def defines_list(self) -> list[str]:
        return self.exp.defines_list

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        assert isinstance(self.exp, Model), f'{self!r}:{self.exp=!r}'
        return self.exp.missing_rules(rulenames)

    def _used_rule_names(self):
        return self.exp._used_rule_names()

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:
        return self.exp._first(k, f)

    def _follow(self, k, fl, a):
        return self.exp._follow(k, fl, a)

    def nodecount(self) -> int:
        return 1 + self.exp.nodecount()

    def _pretty(self, lean=False):
        return self.exp._pretty(lean=lean)

    @cached_property
    def _nullable(self) -> bool:
        return self.exp._nullable

    def optimized(self) -> Self | Model:
        exp = self.exp.optimized()
        new = copy(self)
        new.exp = exp
        return new


@nodedataclass
class Synth(Box):
    """A synthetic placeholder for an unresolved rule."""

    pass


@nodedataclass
class NamedBox(Box):
    name: str = field(default='')  # type: ignore

    def optimized(self) -> Self | Model:
        exp = self.exp.optimized()
        new = copy(self)
        new.exp = exp
        return new


@nodedataclass
class Rule(NamedBox):
    params: tuple[str, ...] = field(default_factory=tuple)
    kwparams: dict[str, Any] = field(default_factory=dict)
    decorators: list[str] = field(default_factory=list)
    base: str | None = None

    # these come from parsing a grammar
    is_name: bool = False
    is_tokn: bool = False
    no_memo: bool = False
    no_stak: bool = False

    # these are used by left recursion analisis
    is_memo: bool = True
    is_lrec: bool = False

    def __post_init__(self):
        super().__post_init__()
        self.params = self.params or ()
        self.kwparams = self.kwparams or {}
        self.decorators = self.decorators or []

        # pyrefly: ignore [unnecessary-type-conversion]
        self.is_name = bool(self.is_name) or 'name' in self.decorators
        # pyrefly: ignore [unnecessary-type-conversion]
        self.is_name = bool(self.is_name) or 'isname' in self.decorators

        if not self.kwparams:
            self.kwparams = {}
        elif not isinstance(self.kwparams, dict):
            self.kwparams = dict(self.kwparams)
        assert isinstance(self.kwparams, dict), f'{typename(self)}: {self.kwparams=!r}'
        self.is_tokn = self.is_tokn or self.name.lstrip('_')[:1].isupper()

    def __pub__(self, sunderok: bool = False) -> dict[str, Any]:
        pub = super().__pub__()
        del pub['exp']
        pub['exp'] = self.exp
        return pub

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return self.exp.missing_rules(rulenames)

    def _parse(self, ctx: Ctx) -> Any:
        return self._parse_rhs(ctx, self.exp)

    def _parse_rhs(self, ctx: Ctx, exp: Model) -> Any:
        ri = RuleInfo.bind(self.ruleinfo, exp, func=exp._parse)
        return ctx.call(ri)

    def _first(self, k, f) -> ffset:
        self._firstset = self.exp._first(k, f) | f[self.name]
        return self._firstset

    def _follow(self, k, fl, a):
        _ = a
        return self.exp._follow(k, fl, fl[self.name])

    @cached_property
    def _nullable(self) -> bool:
        return self.exp._nullable

    @property
    def memoizable(self):
        return self.is_memo and not self.no_memo and not self.is_lrec

    @property
    def should_trace(self) -> bool:
        return not self.no_stak

    @staticmethod
    def param_repr(p):
        if isinstance(p, int | float) or (isinstance(p, str) and p.isalnum()):
            return str(p)
        else:
            return repr(p)

    @property
    def ruleinfo(self):
        if (ri := getattr(self, '_ruleinfo', None)) is None:
            ri = RuleInfo(
                name=self.name,
                instance=self.exp,
                func=self.exp._parse,
                no_memo=self.no_memo,
                no_stak=self.no_stak,
                is_name=self.is_name,
                is_tokn=self.is_tokn,
                params=self.params,
                kwparams=self.kwparams,
                is_lrec=self.is_lrec,
                is_memo=self.memoizable,
            )
        ri = RuleInfo.bind(ri, self, self._parse)
        self._ruleinfo = ri
        return ri

    def _pretty(self, lean=False):
        str_template = "{is_name}{no_memo}{name}{base}{params}:{exp}"

        if lean:
            params = ''
        else:
            params = (
                ', '.join(self.param_repr(p) for p in self.params)
                if self.params
                else ''
            )

            skwparams = ''
            if self.kwparams:
                skwparams = ', '.join(
                    (f'{k}={self.param_repr(v)}' for (k, v) in self.kwparams.items()),
                )

            if params and skwparams:
                params = f'[{params}, {skwparams}]'
            elif skwparams:
                params = f'[{skwparams}]'
            elif params:
                params = f'[{params}]'

        base = f' < {self.base!s}' if self.base else ''

        exp = self.exp._pretty(lean=lean)
        if len(exp.splitlines()) > 1:
            exp = '\n' + indent(exp)
        else:
            exp = ' ' + exp

        return trim(str_template).format(
            name=self.name,
            base=base,
            params=params,
            exp=exp,
            no_memo='@nomemo\n' if self.no_memo else '',
            is_name='@name\n' if self.is_name else '',
        )

    def optimized(self) -> Rule:
        from .syntax import Call, Group, Sequence

        assert isinstance(self.exp, Model)

        exp = self.exp.optimized()
        prev = None
        while exp != prev:
            prev = exp
            if isinstance(exp, Sequence) and len(exp.sequence) == 1:
                exp = exp.sequence[0]

            if isinstance(exp, Call) and exp.rule and not exp.rule.params:
                exp = exp.rule.exp.optimized()

            if isinstance(exp, Group):
                exp = exp.exp.optimized()

        new = copy(self)
        new.exp = exp
        new._lookahead = self.lookahead()
        return new


@nodedataclass
class Patterns(Model):
    whitespace: str | UndefinedType | None = None
    comments: str | None = None
    eol_comments: str | None = None


@nodedataclass
class Grammar(Model):
    name: str = 'MyTest'
    directives: dict[str, Any] = field(default_factory=dict)
    keywords: tuple[str, ...] = field(default_factory=tuple)
    rules: tuple[Rule, ...] = field(default_factory=tuple)
    _optimized: Grammar | None = None

    def __init__(
        self,
        name=None,
        rules: Iterable[Rule] | None = None,
        *,
        config: Any = None,
        directives: dict | None = None,
        keywords: tuple[str, ...] | None = None,
        **settings: Any,
    ):
        super().__init__()
        config = config if isinstance(config, ParserConfig) else ParserConfig()
        assert isinstance(rules, Iterable), f'{type(rules)!r} {rules!r}'

        directives = directives or {}
        assert isinstance(directives, dict)
        self.directives = directives

        config = ParserConfig.new(config=config, **settings)
        config = config.hard_override(**directives)
        self._config: ParserConfig = config

        keywords = keywords or config.keywords or ()
        keywords = tuple(k for k in keywords or () if k)
        assert isinstance(keywords, tuple)
        if self.config.ignorecase:
            keywords = tuple(k.upper() for k in keywords if k)
        keywords = tuple(sorted(set(keywords)))
        assert isinstance(keywords, tuple)
        self.keywords = keywords

        # NOTE Ctx needs keywords in config
        self._config = self._config.override(keywords=self.keywords)

        self.rules = tuple(rules)  # type: ignore

        self.name = self._resolve_name(name)
        self.initialize()

    def initialize(self) -> None:
        rulemap = {rule.name: rule for rule in self.rules}
        self._rule = SimpleNamespace(**rulemap)
        self._rulemap = self._rule.__dict__
        self.link(self)
        self._calc_lookahead_sets()
        self._mark_left_recursion()

        missing: set[str] = self.missing_rules(set(self.rulemap))
        if missing:
            msg = ' '.join(missing)
            raise GrammarError('unknown rules, no parser generated: ' + msg)

    def configure(self, config: ParserConfig | None = None, **settings: Any):
        self._config.merge_config(config)
        self._config.merge(**settings)

    def _update_patterns(self):
        if not hasattr(self, 'patterns'):
            self.patterns = Patterns()
        self.patterns.whitespace = (
            self.patterns.whitespace
            or self.config.whitespace
            or self.directives.get('whitespace')
        )
        self.patterns.comments = (
            self.patterns.comments
            or self.config.comments
            or self.directives.get('comments')
        )
        self.patterns.eol_comments = (
            self.patterns.eol_comments
            or self.config.eol_comments
            or self.directives.get('eol_comments')
        )

    @property
    def config(self) -> ParserConfig:
        return self._config

    @property
    def rulemap(self) -> dict[str, Rule]:
        return self._rulemap

    @property
    def rule(self) -> SimpleNamespace:
        return self._rule

    @property
    def semantics(self) -> Any:
        return self.config.semantics

    @semantics.setter
    def semantics(self, value: Any):
        self.config.semantics = value  # type: ignore

    @property
    def first_sets(self):
        return self._firstset

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set().union(*[rule.missing_rules(rulenames) for rule in self.rules])

    def used_rules(self) -> list[Rule]:
        used = self._used_rule_names()
        return [rule for rule in self.rules if rule.name in used]

    def _resolve_name(self, name: str | None) -> str:
        if name is None:
            name = self.directives.get('grammar')
        if name is None:
            name = self.config.name
        if name is None and self.config.source is not None:
            name = Path(self.config.source).stem
        if name is None:
            name = 'My'
        return name

    def _mark_left_recursion(self):
        # from .leftrec.depthf import mark_left_recursion
        # from .leftrec.autumn import mark_left_recursion
        from .leftrec.pegen import mark_left_recursion

        leftrect_rules = mark_left_recursion(self.rules)
        if leftrect_rules and not self.config.left_recursion:
            config = self.config
            raise GrammarError(
                f'{config.left_recursion=}'
                f' but found left-recursive rules'
                f' {', '.join(repr(r.name) for r in leftrect_rules)}!',
            )

    def _used_rule_names(self) -> set[str]:
        if not self.rules:
            return set()

        used = {'start', self.rules[0].name}
        prev: set[str] = set()
        while used != prev:
            prev = used
            used |= set().union(
                *[rule._used_rule_names() for rule in self.rules if rule.name in used],
            )
        return used

    def _calc_lookahead_sets(self, k: int = 1):
        self._calc_first_sets(k=k)
        self._calc_follow_sets(k=k)
        for rule in self.rules:
            if rule.lookaheadlist is None:
                raise GrammarError(f'Impossible lookahead in rule {rule.name}')

    def _calc_first_sets(self, k: int = 1):
        f: dict[str, ffset] = defaultdict(set)
        f1 = None
        while f1 != f:
            f1 = copy(f)
            for rule in self.rules:
                f[rule.name] |= rule._first(k, f)

        # cache results
        for rule in self.rules:
            rule._firstset = f[rule.name]

    def _calc_follow_sets(self, k: int = 1):
        fl: dict[str, ffset] = defaultdict(set)
        fl1 = None
        while fl1 != fl:
            fl1 = copy(fl)
            for rule in self.rules:
                rule._follow(k, fl, set())

        # cache results
        for rule in self.rules:
            rule._follow_set = fl[rule.name]

    def parse(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> Any:
        self = self.optimized()  # noqa: PLW0642
        return self._do_parse(
            text, start=start, config=config, asmodel=asmodel, **settings
        )

    def _do_parse(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> Any:
        config = self.new_parse_config(start=start, config=config, **settings)
        ctx = self.newctx(asmodel=asmodel)
        return ctx.parse(text, config=config)

    def newctx(self, asmodel: bool = True) -> Ctx:
        return ModelContext(self.rules, config=self.config, asmodel=asmodel)

    def new_parse_config(
        self,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        **settings: Any,
    ) -> Any:
        config = self.config.override_config(config)
        assert isinstance(config, ParserConfig)
        # NOTE: bw-comp: allow overriding directives
        config = config.override(start=start, **settings)

        if isinstance(config.semantics, type):
            raise TypeError(
                'semantics must be an object instance or None, '
                f'not class {config.semantics!r}',
            )

        start = config.effective_start_rule_name()
        if start is None:
            start = self.rules[0].name
            config = config.override(
                start=start,
                start_rule=None,
                rule_name=None,
            )
        return config

    def nodecount(self) -> int:
        return 1 + sum(r.nodecount() for r in self.rules)

    def _pretty(self, lean: bool = False) -> str:
        regex_directives = {'comments', 'eol_comments', 'whitespace'}
        string_directives = {'namechars'}

        directives = ''
        # noinspection PyUnresolvedReferences
        for name, value in self.directives.items():
            if name in regex_directives:
                if '/' in value:
                    directives += f'@@{name} :: ?"{value}"\n'
                else:
                    directives += f'@@{name} :: /{value}/\n'
            elif name in string_directives:
                directives += f'@@{name} :: {value!r}\n'
            else:
                directives += f'@@{name} :: {value}\n'
        if directives:
            directives += '\n'

        keywordsets = []
        batch: list[str] = []
        bfmt: str = ""
        for k in sorted(repr(k) for k in self.keywords):
            batch += [k]
            bfmt = f"@@keyword :: {' '.join(batch)}"
            if len(bfmt) >= PEP8_LLEN - 8:
                keywordsets += [bfmt]
                batch = []
        if batch:
            keywordsets += [bfmt]

        keywords = f"\n\n{'\n'.join(keywordsets)}\n\n" if keywordsets else ""

        rules = (
            '\n\n'.join(str(rule._pretty(lean=lean)) for rule in self.rules)
        ).rstrip() + '\n\n'
        return f"{directives}{keywords}{rules}"

    def optimized(self) -> Grammar:
        if isinstance(self._optimized, Grammar):
            return self._optimized

        optrules: tuple[Rule, ...] = tuple(r.optimized() for r in self.rules)
        new = copy(self)
        new.rules = optrules
        new.initialize()

        self._optimized = new  # NOTE cache optimized grammar
        new._optimized = new  # NOTE circular reference as cached

        return new

    @classmethod
    def __from_json__(cls: type[Self], data: Mapping[str, Any]) -> Grammar:
        g = super().__from_json__(data)
        # NOTE got a Grammar that used Grammar.__init__()
        #
        # NOTE
        #   Copy the grammar so the configuration is unique
        # HACK
        return Grammar(
            name=g.name,
            rules=g.rules,
            directives=g.directives,
            keywords=g.keywords,
        )


class ModelContext(ParseContext):
    def __init__(
        self,
        rules: Iterable[Rule],
        /,
        start: str | None = None,
        config: ParserConfig | None = None,
        asmodel: bool = False,
        **settings: Any,
    ):
        config = ParserConfig.new(config, **settings)
        assert isinstance(config, ParserConfig)
        super().__init__(start=start, config=config, asmodel=asmodel)
        if not self.config.semantics and asmodel:
            self.config.semantics = ModelBuilderSemantics()

        self._rulemap: dict[str, Rule] = {rule.name: rule for rule in rules}

    @property
    def rulemap(self) -> dict[str, Rule]:
        return self._rulemap

    def find_rule(self, name: str) -> Func:
        return self.rulemap[name]._parse



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
Visual verification tables for the ANSI colour library.

.. code-block:: text

    python -m tatsu.util.ztyle

Displays the 16x16 ANSI colour cube, the named colour palette,
dynamic colour-method tests, and the CSS named colour table.
"""

from __future__ import annotations

from pathlib import Path

from .colormap import COLORMAP
from .csscolormap import CSS_COLORS
from .style import Style


def main():
    """Print all verification tables to stdout."""
    print_color_table()
    print_colormap_table()
    print_color_method_table()
    print_css_color_table()
    # add_color_methods()

    print(
        f'{Style(value="A pink", fg=200)} and {Style(value="a magenta", fg=201)} {Style(value="aquamarine", fg=72)} world'
    )


def col_header(row: int, width: int = 4):
    """Print a row-number header cell, right-aligned in *width* columns."""
    print(f'{row if row >= 0 else "":>{width}}', end=" ")


def header(width: int = 4):
    """Print the 16-column colour-table header."""
    col_header(-1, width)
    for col in range(16):
        print(f'{col:>{width}d}', end=" ")


def print_color_table():
    """Print the 16x16 ANSI colour cube (0-255) as a grid."""
    header()
    for row in range(16):
        col_header(row, 4)
        for col in range(16):
            index = row * 16 + col
            print(Style(value=f'{index:>3d}', fg=index), end=" ")
        print()
    print()


def print_colormap_table():
    """Print all named ANSI colours in their own foreground colour."""
    width = 24
    k = 100 // width
    for n, name in enumerate(COLORMAP.keys()):
        col = n % k
        if col == 0:
            print()
        print(Style(value=f'{name:>{width}s}', fg=COLORMAP[name]), end=" ")
    print()
    print()


def print_color_method_table():
    """Print the dynamic colour-method names in their own foreground."""
    width = 24
    k = 100 // width
    n = 0
    for n, name in enumerate(COLORMAP):
        style = Style(f'{name:>{width}s}')
        m = getattr(style, name.replace(' ', '_'), None)
        if m is not None:
            style = m()

        col = n % k
        if col == 0:
            print()
        print(style, end=" ")
    print()
    print()


def print_css_color_table():
    """Print all CSS named colours swatched in their own foreground."""
    width = 24
    k = 100 // width
    for n, (name, rgb) in enumerate(CSS_COLORS.items()):
        col = n % k
        if col == 0:
            print()
        swatch = Style(value=f'{name:>{width}s}', fg=rgb)
        print(swatch, end=" ")
    print()
    print()


def add_color_methods():
    cls = Style
    with Path("./methods.py").open("tw+") as f:
        for name, code in COLORMAP.items():
            method_name = name.replace(' ', '_')
            if hasattr(Style, method_name):
                continue

            def _color_method(self, s: str = "", code: int = code) -> Style:
                return type(self)(value=s, fg=code)

            setattr(
                cls,
                method_name,
                _color_method,
            )

    with Path("./methods.py").open("tw+") as f:
        for name, code in sorted(COLORMAP.items()):
            method_name = name.replace(' ', '_')
            f.write(f'    def {method_name}(self) -> Self:\n')
            f.write(f'        return self.fg({code})\n')
            f.write('\n')

            f.write(f'    def {method_name}_bg(self) -> Self:\n')
            f.write(f'        return self.bg({code})\n')
            f.write('\n')

            f.flush()


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
from functools import cached_property
from typing import Any

from ..contexts import Ctx
from ..objectmodel import nodedataclass
from ..util import regexpp, trim
from .base import Leaf
from .math import ffset


@nodedataclass
class Pattern(Leaf):
    pattern: str = ""

    def __post_init__(self):
        super().__post_init__()
        self.pattern = self.pattern or self.ast or ""
        try:
            self._regex = re.compile(self.pattern)
        except Exception as e:
            raise ValueError(f"Invalid regex pattern: {self.pattern!r}") from e

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.pattern(self.pattern or r'\\')

    def _first(self, k, f) -> ffset:
        _ = k, f
        x = str(self)
        if bool(self._regex.match('')):
            return {(), (x,)}
        else:
            return {(x,)}

    def _pretty(self, lean=False):
        _ = lean
        pat = self.pattern or ""
        # multiline patterns are OK
        pat = trim(pat)
        if '/' in pat:
            newpat = pat.replace('"', r'\"')
            regex = f'?"{newpat}"'
        else:
            regex = f'/{pat}/'
        return regex

    @cached_property
    def _nullable(self) -> bool:
        return bool(self._regex.match(''))

    def __str__(self) -> str:
        return regexpp(self.pattern)[2:-1]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .grammar_gen import parsermodel_gen as parsergen
from .ngmodel_gen import modelgen
from .ngparser_gen import pythongen, pythongen as codegen


__all__ = ['codegen', 'parsergen', 'pythongen', 'modelgen']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations


_PYGMENTS_AVAILABLE: bool | None = None


def is_pygments_available() -> bool:
    global _PYGMENTS_AVAILABLE  # noqa: PLW0603
    if _PYGMENTS_AVAILABLE is None:
        try:
            import pygments  # noqa: F401
        except ImportError:
            _PYGMENTS_AVAILABLE = False
        else:
            _PYGMENTS_AVAILABLE = True
    return _PYGMENTS_AVAILABLE



import asyncio
import sys


class MultiProgressBar:
    def __init__(self, total_bars: int):
        # State tracking for all bars
        self.bars = [
            {"current": 0, "total": 100, "label": f"Task {i}"}
            for i in range(total_bars)
        ]
        self.running = True

    def update(self, bar_index: int, value: int):
        """Workers call this to increment their progress."""
        self.bars[bar_index]["current"] = min(value, self.bars[bar_index]["total"])

    async def render_loop(self):
        """The single task that handles terminal I/O rendering."""
        # Hide the terminal cursor
        sys.stdout.write("\033[?25l")
        sys.stdout.flush()

        try:
            while self.running:
                output = []
                for bar in self.bars:
                    pct = int((bar["current"] / bar["total"]) * 100)
                    filled = int(pct / 5)  # 20 blocks total
                    blocks = "█" * filled + "-" * (20 - filled)
                    output.append(f"{bar['label']}: [{blocks}] {pct}%")

                # Join with newlines
                full_screen = "\n".join(output)
                sys.stdout.write(full_screen)
                sys.stdout.flush()

                await asyncio.sleep(0.05)  # ~20 FPS refresh rate

                # Move cursor back up to the top of our progress bars
                sys.stdout.write(f"\033[{len(self.bars)}A")
        finally:
            # Clean up: Move cursor past the bars and show it again
            sys.stdout.write(f"\033[{len(self.bars)}B\n\033[?25h")
            sys.stdout.flush()


# --- Simulation of Parallel Workers ---


async def worker(bar_id: int, manager: MultiProgressBar):
    """Simulates an I/O bound worker updating its progress."""
    for i in range(1, 101):
        await asyncio.sleep(0.02 + (bar_id * 0.01))  # Staggered speeds
        manager.update(bar_id, i)


async def main():
    num_tasks = 4
    manager = MultiProgressBar(num_tasks)

    # Start the single rendering loop task
    renderer_task = asyncio.create_task(manager.render_loop())

    # Start all parallel workers
    workers = [worker(i, manager) for i in range(num_tasks)]
    await asyncio.gather(*workers)

    # Stop the renderer once workers finish
    manager.running = False
    await renderer_task


if __name__ == "__main__":
    asyncio.run(main())



import os
import sys
from dataclasses import dataclass, field
from pathlib import Path


DEFAULT_PYGMENTS_STYLE = "nord"


class CLIError(Exception):
    """Raised when the CLI encounters an error."""


@dataclass
class CLIConfig:
    """Parsed command-line configuration, matching ogopego's CLIConfig struct."""

    # Global flags
    color: str = "auto"
    output: str | None = None
    style: str = DEFAULT_PYGMENTS_STYLE
    verbose: bool = False
    quiet: bool = False
    profile: bool = False
    trace: bool = False

    # Subcommand state
    command: str = ""
    grammar: str | Path = ""
    optimized: bool = False
    path: str | None = None
    inputs: list[str] = field(default_factory=list)

    # format flags
    json: bool = False
    json_lines: bool = False
    model: bool = False
    pretty: bool = False
    pretty_lean: bool = False
    railroads: bool = False
    object_model: bool = False
    parser_model: bool = False
    generate_parser: bool = False

    # run flags
    start: str | None = None
    name: str = ""
    nproc: int | None = None
    summary: bool = False

    @property
    def usecolor(self) -> bool:
        return self.color == "always" or (  # pyright: ignore[reportReturnType]
            not bool(os.environ.get("NO_COLOR", None))
            and self.color == "auto"
            and sys.stdout.isatty()
        )



@dataclass(slots=True, kw_only=True)
class ColorBar(Bar):
    color_code: str = "\033[32m"  # Default to Green ANSI

    def bar(self) -> list[tuple[str, int]]:
        pct = int((self.current / self.total) * 100) if self.total > 0 else 0
        fill = int(pct / 5)

        # Build chunks: (String with ANSI controls, actual screen character length)
        return [
            (f"{self.label}: [", len(self.label) + 4),
            (f"{self.color_code}{'█' * fill}\033[0m", fill),
            (f"{'-' * (20 - fill)}] {pct}%", (20 - fill) + 2 + len(str(pct)) + 1),
        ]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import hashlib
import os
import shutil
import string
import time
from typing import Any

from .itertools import first  # noqa  # pyright: ignore[reportUnusedImport]


ALPHABET = string.digits + string.ascii_uppercase
GREEKTOME = "αβδεζηθικλμνξοπρστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ"
GREEKtome = "αβδεζηθικλμνξοπρστυφχψω"


def is_main_process() -> bool:
    import multiprocessing

    return multiprocessing.parent_process() is None


MAIN_PID = os.getpid()


def check_pid() -> bool:
    return os.getpid() == MAIN_PID


def platform_has_command(name) -> bool:
    # noinspection PyDeprecation
    return shutil.which(name) is not None


def clock_time_μs() -> int:
    # WARNING Time is a crucial concept when dealing with concurrency
    # WARNING and wanting to calculate elapsed time accurately.
    return time.clock_gettime_ns(time.CLOCK_REALTIME) // 1000


def ascii2greek(value: str) -> str:
    a = string.ascii_lowercase
    m = max(len(a), len(GREEKTOME))
    g = (GREEKTOME * 2)[:m]

    return value.translate(str.maketrans(a, g, string.punctuation + string.whitespace))


def greek2ascii(value: str) -> str:
    a = string.ascii_lowercase
    g = GREEKTOME
    return value.translate(str.maketrans(g, a * 2))


def i2alpha(value: int, width: int = 1, alphabet: str = ALPHABET):
    value = int(value)
    width = int(width)
    alphabet = str(alphabet)

    if value == 0:
        return alphabet[0] * width

    n = len(alphabet)

    def gen(value: int):
        while value != 0:
            value, c = divmod(value, n)
            yield alphabet[c]

    result = "".join(reversed(list(gen(value))))
    return result.rjust(width, alphabet[0])


def i2Greek(value: int, width: int = 1):
    return i2alpha(value, width, GREEKTOME)


def i2greek(value: int, width: int = 1):
    return i2alpha(value, width, GREEKtome)


def greek_time() -> str:
    return i2greek(int(time.time()), width=8)


def alpha_timestamp() -> str:
    t = time.time()
    y, d = divmod(t, 24 * 3600)
    h, s = divmod(d, 3600)
    m, _ = divmod(s, 0.36)
    return f"{i2alpha(int(y), alphabet=string.ascii_lowercase)}-{h:02.0f}-{m:04.0f}"


def lower_time() -> str:
    return i2alpha(int(time.time()), width=8, alphabet=string.ascii_lowercase)


def new_id() -> str:
    d = 8
    t = time.monotonic_ns()
    _mm, mn = divmod(t, 10**d)
    return i2greek(mn, width=d)


def hash2byte(data: Any) -> bytes:
    if data is None:
        return b"\x00" * 4
    return hashlib.blake2b(str(data).encode("utf-8"), digest_size=2).digest()


def hash2str(data: Any) -> str:
    return f"{int.from_bytes(hash2byte(data), 'big'):04x}"



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .log import iso_logpath, logctx, startscript


__all__ = [
    "iso_logpath",
    "logctx",
    "startscript",
]



import copy
import threading
import time


class MultiProgress:
    def __init__(self):
        self._bars = []
        self._lock = threading.Lock()
        self._running = False
        self._thread = None

    def add_bar(self, bar_instance: Bar) -> Bar:
        """Accepts any pre-configured Bar or custom subclass instance."""
        with self._lock:
            self._bars.append(bar_instance)
        return bar_instance

    def start(self):
        self._running = True
        sys.stdout.write("\033[?25l")
        sys.stdout.flush()
        self._thread = threading.Thread(target=self._render_loop, daemon=True)
        self._thread.start()

    def stop(self):
        self._running = False
        if self._thread:
            self._thread.join()
        with self._lock:
            num_bars = len(self._bars)
        sys.stdout.write(f"\033[{num_bars}B\n\033[?25h")
        sys.stdout.flush()

    def _render_loop(self):
        while self._running:
            with self._lock:
                snapshot = copy.deepcopy(self._bars)

            if not snapshot:
                time.sleep(0.05)
                continue

            lines = []
            for b in snapshot:
                # Ask the data object to render its chunks
                chunks = b.bar()
                # Join the actual text payload together
                line_string = "".join(text for text, _ in chunks)
                lines.append(line_string)

            sys.stdout.write("\n".join(lines))
            sys.stdout.flush()

            time.sleep(0.05)  # 20 FPS
            sys.stdout.write(f"\033[{len(snapshot)}A")



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Callable, Generator, Iterable
from pathlib import Path
from typing import Any

from ..packetz import PacketLike
from ..util import identity
from .parproc import parproc
from .task import Func
from .visual import VisualPayload, parproc_visual


def parallel_proc(
    payloads: Iterable[Any],
    process: Callable,
    /,
    *args: Any,
    pickable: Func = identity,
    parallel: bool = True,
    reraise: bool = False,
    **kwargs: Any,
) -> Generator[PacketLike | None, None, None]:
    yield from parproc(
        process,
        payloads,
        *args,
        pickable=pickable,
        parallel=parallel,
        reraise=reraise,
        **kwargs,
    )


def processing_loop(
    filenames: Iterable[str],
    process: Callable,
    /,
    *args: Any,
    pickable: Func = identity,
    parallel: bool = True,
    reraise: bool = False,
    max_workers: int | None = None,
    **kwargs: Any,
) -> Generator[PacketLike, None, None]:
    paths = [Path(f) for f in filenames]
    payloads = [VisualPayload(p, p.read_text()) for p in paths]
    yield from parproc_visual(
        process,
        payloads,
        *args,
        pickable=pickable,
        parallel=parallel,
        reraise=reraise,
        max_workers=max_workers,
        **kwargs,
    )
